import numpy as np

from foxes.core import TurbineModel
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC


class SetFarmVars(TurbineModel):
    """
    Set farm data variables to given data.

    Attributes
    ----------
    vars: list of str
        The variables to be set
    once: bool
        Flag for running only once

    :group: models.turbine_models

    """

    def __init__(self, pre_rotor=False, once=False):
        """
        Constructor.

        Parameters
        ----------
        pre_rotor: bool
            Flag for running this model before
            running the rotor model.
        once: bool
            Flag for running only once

        """
        super().__init__(pre_rotor=pre_rotor)
        self.once = once
        self.reset()

    def add_var(self, var, data):
        """
        Add data for a variable.

        Parameters
        ----------
        var: str
            The variable name
        data: numpy.ndarray
            The data, shape: (n_states, n_turbines)

        """
        if self.initialized:
            raise ValueError(
                f"Model '{self.name}': Cannot add_var after initialization"
            )
        if self.running:
            raise ValueError(f"Model '{self.name}': Cannot add_var while running")
        self.vars.append(var)
        self.__vdata.append(np.asarray(data, dtype=config.dtype_double))

    def reset(self):
        """
        Remove all variables.
        """
        if self.running:
            raise ValueError(f"Model '{self.name}': Cannot reset while running")
        self.vars = []
        self.__vdata = []

    def initialize(self, algo, verbosity=0, force=False):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent
        force: bool
            Overwrite existing data

        """
        super().initialize(algo, verbosity, force)
        self.__once_done = set()

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.vars

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        idata = super().load_data(algo, verbosity)

        for i, v in enumerate(self.vars):
            data = np.full(
                (algo.n_states, algo.n_turbines), np.nan, dtype=config.dtype_double
            )
            vdata = self.__vdata[i]

            # handle special case of call during vectorized optimization:
            if (
                np.ndim(vdata)
                and vdata.shape[0] != algo.n_states
                and hasattr(algo.states, "n_pop")
            ):
                n_pop = algo.states.n_pop
                n_ost = algo.states.states.size()
                n_trb = algo.n_turbines
                vdata = np.zeros((n_pop, n_ost, n_trb), dtype=config.dtype_double)
                vdata[:] = self.__vdata[i][None, :]
                vdata = vdata.reshape(n_pop * n_ost, n_trb)

            data[:] = vdata
            idata["data_vars"][self.var(v)] = ((FC.STATE, FC.TURBINE), data)

            # special case of turbine positions:
            if v in [FV.X, FV.Y]:
                i = [FV.X, FV.Y].index(v)
                for ti in range(algo.n_turbines):
                    t = algo.farm.turbines[ti]
                    if len(t.xy.shape) == 1:
                        xy = np.zeros((algo.n_states, 2), dtype=config.dtype_double)
                        xy[:] = t.xy[None, :]
                        t.xy = xy
                    t.xy[:, i] = np.where(
                        np.isnan(data[:, ti]), t.xy[:, i], data[:, ti]
                    )

            # special case of rotor diameter and hub height:
            if v in [FV.D, FV.H]:
                for ti in range(algo.n_turbines):
                    t = algo.farm.turbines[ti]
                    x = np.zeros(algo.n_states, dtype=config.dtype_double)
                    if v == FV.D:
                        x[:] = t.D
                        t.D = x
                    else:
                        x[:] = t.H
                        t.H = x
                    x[:] = np.where(np.isnan(data[:, ti]), x, data[:, ti])

        return idata

    def set_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to running, and moves
        all large data to stash.

        The stashed data will be returned by the
        unset_running() function after running calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Large data stash, this function adds data here, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().set_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None:
            data_stash[self.name]["vdata"] = self.__vdata
        del self.__vdata

    def unset_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to not running, recovering large data
        from stash

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Reconstruct model data from this stash, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().unset_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None:
            self.__vdata = data_stash[self.name].pop("vdata")

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        self.ensure_output_vars(algo, fdata)

        i0 = mdata.states_i0(counter=True)
        if not self.once or i0 not in self.__once_done:
            bsel = np.zeros((fdata.n_states, fdata.n_turbines), dtype=bool)
            bsel[st_sel] = True

            for v in self.vars:
                data = mdata[self.var(v)]
                hsel = ~np.isnan(data)
                tsel = bsel & hsel

                fdata[v][tsel] = data[tsel]

            self.__once_done.add(i0)

        return {v: fdata[v] for v in self.vars}



import numpy as np
import pandas as pd
from xarray import Dataset
from numbers import Number

from foxes.core import States
from foxes.utils import Dict
import foxes.variables as FV
import foxes.constants as FC

from .read_fields import read_wind_resource_field


default_values = {
    FV.WS: 8.0,
    FV.WD: 270.0,
    FV.TI: 0.1,
    FV.RHO: 1.225,
}


def _get_profiles(coords, fields, dims, ovars, fixval, verbosity):
    """Read ABL profiles information
    :group: input.yaml.windio
    """
    profiles = {}
    if FV.Z0 in fields:
        if FV.H not in fields:
            if verbosity > 0:
                print(
                    f"Ignoring '{FV.Z0}', since no reference_height found. No ABL profile activated."
                )
            fields.pop(FV.Z0)
            dims.pop(FV.Z0)
        elif FV.MOL in fields:
            ovars.append(FV.MOL)
            fixval[FV.H] = fields[FV.H]
            profiles = {FV.WS: "ABLLogWsProfile"}
        else:
            fixval[FV.H] = fields[FV.H]
            profiles = {FV.WS: "ABLLogNeutralWsProfile"}
    elif FV.H in fields and verbosity > 0:
        print(
            f"Ignoring '{FV.H}', since no '{FV.Z0}' data found. No ABL profile activated."
        )
    if len(profiles) and verbosity > 2:
        print(
            f"        Selecting ABL profile '{profiles[FV.WS]}', {FV.H} = {fields[FV.H]} m"
        )

    return profiles


def _get_SingleStateStates(
    coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
):
    """Try to generate single state parameters
    :group: input.yaml.windio
    """
    for c in coords:
        if not isinstance(c, Number):
            return False

    if verbosity > 2:
        print("        selecting class 'SingleStateStates'")

    smap = {FV.WS: "ws", FV.WD: "wd", FV.TI: "ti", FV.RHO: "rho"}

    data = {smap[v]: fixval.get(v, default_values[v]) for v in ovars}
    for v, d in coords.items():
        if v in smap:
            data[smap[v]] = d
        elif verbosity > 1:
            print(f"        ignoring coord '{v}'")
    for v, d in fields.items():
        if v in smap and len(dims[v]) == 0:
            data[smap[v]] = d
        elif verbosity > 1:
            print(f"        ignoring field '{v}' with dims {dims[v]}")

    sdata = pd.DataFrame(index=coords[FC.TIME], data=data)
    sdata.index.name = FC.TIME
    states_dict.update(
        dict(
            states_type="SingleStateStates",
            profiles=profiles,
            **data,
        )
    )
    return True


def _get_Timeseries(
    coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
):
    """Try to generate time series parameters
    :group: input.yaml.windio
    """
    if len(coords) == 1 and FC.TIME in coords:
        if verbosity > 2:
            print("        selecting class 'Timeseries'")

        data = {}
        fix = {v: fixval.get(v, default_values[v]) for v in ovars if v not in fields}
        for v, d in fields.items():
            if dims[v] == (FC.TIME,):
                data[v] = d
            elif len(dims[v]) == 0:
                fix[v] = d
            elif verbosity > 2:
                print(f"        ignoring field '{v}' with dims {dims[v]}")

        sdata = pd.DataFrame(index=coords[FC.TIME], data=data)
        sdata.index.name = FC.TIME
        states_dict.update(
            dict(
                states_type="Timeseries",
                data_source=sdata,
                output_vars=ovars,
                fixed_vars=fix,
                profiles=profiles,
            )
        )
        return True
    return False


def _get_MultiHeightNCTimeseries(
    coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
):
    """Try to generate time series parameters
    :group: input.yaml.windio
    """
    if len(coords) == 2 and FC.TIME in coords and FV.H in coords:
        if verbosity > 2:
            print("        selecting class 'MultiHeightNCTimeseries'")

        if len(profiles) and verbosity > 0:
            print(
                f"Ignoring profile '{profiles[FV.WS]}' for states class 'MultiHeightNCTimeseries'"
            )

        data = {}
        fix = {v: fixval.get(v, default_values[v]) for v in ovars if v not in fields}
        for v, d in fields.items():
            if dims[v] == (FC.TIME, FV.H):
                data[v] = ((FC.TIME, FV.H), d)
            elif dims[v] == (FV.H, FC.TIME):
                data[v] = ((FC.TIME, FV.H), np.swapaxes(d, 0, 1))
            elif len(dims[v]) == 0:
                fix[v] = d
            elif verbosity > 2:
                print(f"        ignoring field '{v}' with dims {dims[v]}")

        sdata = Dataset(coords=coords, data_vars=data)
        states_dict.update(
            dict(
                states_type="MultiHeightNCTimeseries",
                h_coord=FV.H,
                format_times_func=None,
                data_source=sdata,
                output_vars=ovars,
                fixed_vars=fix,
                bounds_error=False,
            )
        )
        return True
    return False


def _get_WeibullSectors(
    coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
):
    """Try to generate Weibull sector parameters
    :group: input.yaml.windio
    """
    if (
        FV.WEIBULL_A in fields
        and FV.WEIBULL_k in fields
        and "sector_probability" in fields
        and len(dims[FV.WEIBULL_A]) == 1
        and len(dims[FV.WEIBULL_k]) == 1
        and len(dims["sector_probability"]) == 1
    ):
        if verbosity > 2:
            print("        selecting class 'WeibullSectors'")

        data = {}
        fix = {v: fixval.get(v, default_values[v]) for v in ovars if v not in fields}
        c = dims[FV.WEIBULL_A][0]
        for v, d in fields.items():
            if dims[v] == (c,):
                data[v] = d
            elif len(dims[v]) == 0:
                fix[v] = d
            elif verbosity > 2:
                print(f"        ignoring field '{v}' with dims {dims[v]}")

        if FV.WD in coords:
            data[FV.WD] = coords[FV.WD]

        sdata = pd.DataFrame(index=range(len(fields[FV.WEIBULL_A])), data=data)
        sdata.index.name = "sector"
        states_dict.update(
            dict(
                states_type="WeibullSectors",
                data_source=sdata,
                ws_bins=np.arange(60) / 2 if FV.WS not in sdata else None,
                output_vars=ovars,
                var2ncvar={FV.WEIGHT: "sector_probability"},
                fixed_vars=fix,
                profiles=profiles,
            )
        )
        return True
    return False


def _get_WeibullPointCloud(
    coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
):
    """Try to generate Weibull sector parameters
    :group: input.yaml.windio
    """
    if (
        FV.WD in coords
        and FV.WEIBULL_A in fields
        and FV.WEIBULL_k in fields
        and "sector_probability" in fields
        and FV.X in fields
        and FV.Y in fields
        and len(dims[FV.X]) == 1
        and dims[FV.X] == dims[FV.Y]
        and dims[FV.X][0] != FV.WD
        and dims[FV.X][0] in coords
    ):
        if verbosity > 2:
            print("        selecting class 'WeibullPointCloud'")

        data = {}
        fix = {v: fixval.get(v, default_values[v]) for v in ovars if v not in fields}
        for v, d in fields.items():
            if len(dims[v]) == 0:
                fix[v] = d
            elif v not in fixval:
                data[v] = (dims[v], d)

        sdata = Dataset(
            coords=coords,
            data_vars=data,
        )

        states_dict.update(
            dict(
                states_type="WeibullPointCloud",
                data_source=sdata,
                output_vars=ovars,
                var2ncvar={},
                fixed_vars=fix,
                point_coord=dims[FV.X][0],
                wd_coord=FV.WD,
                ws_coord=FV.WS if FV.WS in sdata.coords else None,
                ws_bins=np.arange(60) / 2 if FV.WS not in sdata else None,
                x_ncvar=FV.X,
                y_ncvar=FV.Y,
                h_ncvar=FV.H if FV.H in sdata.data_vars else None,
                weight_ncvar="sector_probability",
            )
        )
        return True
    return False


def _get_WeibullField(
    coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
):
    """Try to generate Weibull sector parameters
    :group: input.yaml.windio
    """
    if (
        FV.WD in coords
        and FV.X in coords
        and FV.Y in coords
        and FV.WEIBULL_A in fields
        and FV.WEIBULL_k in fields
        and "sector_probability" in fields
    ):
        if verbosity > 2:
            print("        selecting class 'WeibullField'")

        data = {}
        fix = {v: fixval.get(v, default_values[v]) for v in ovars if v not in fields}
        for v, d in fields.items():
            if len(dims[v]) == 0:
                fix[v] = d
            elif v not in fixval:
                data[v] = (dims[v], d)

        sdata = Dataset(
            coords=coords,
            data_vars=data,
        )

        states_dict.update(
            dict(
                states_type="WeibullField",
                data_source=sdata,
                output_vars=ovars,
                wd_coord=FV.WD,
                x_coord=FV.X,
                y_coord=FV.Y,
                h_coord=FV.H if FV.H in sdata.coords else None,
                weight_ncvar="sector_probability",
                var2ncvar={},
                fixed_vars=fix,
                ws_bins=np.arange(60) / 2 if FV.WS not in sdata.coords else None,
                ws_coord=FV.WS if FV.WS in sdata.coords else None,
            )
        )
        return True
    return False


def get_states(coords, fields, dims, verbosity=1):
    """
    Reads states parameters from windio input

    Parameters
    ----------
    coords: dict
        The coordinates data
    fields: dict
        The fields data
    dims: dict
        The dimensions data
    verbosity: int
        The verbosity level

    Returns
    -------
    states: foxes.core.States
        The states object

    :group: input.yaml.windio

    """
    if verbosity > 2:
        print("      Creating states")

    ovars = [FV.WS, FV.WD, FV.TI, FV.RHO]
    fixval = {}
    profiles = _get_profiles(coords, fields, dims, ovars, fixval, verbosity)

    states_dict = {}
    if (
        _get_SingleStateStates(
            coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
        )
        or _get_Timeseries(
            coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
        )
        or _get_MultiHeightNCTimeseries(
            coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
        )
        or _get_WeibullPointCloud(
            coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
        )
        or _get_WeibullSectors(
            coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
        )
        or _get_WeibullField(
            coords, fields, dims, states_dict, ovars, fixval, profiles, verbosity
        )
    ):
        return States.new(**states_dict)
    else:
        raise ValueError(
            f"Failed to create states for coords {list(coords.keys())} and fields {list(fields.keys())} with dims {dims}"
        )


def read_site(wio_dict, verbosity=1):
    """
    Reads the site information

    Parameters
    ----------
    wio_dict: foxes.utils.Dict
        The windio data
    verbosity: int
        The verbosity level, 0=silent

    Returns
    -------
    states: foxes.core.States
        The states object

    :group: input.yaml.windio

    """

    def _print(*args, level=1, **kwargs):
        if verbosity >= level:
            print(*args, **kwargs)

    wio_site = wio_dict["site"]
    _print("Reading site")
    _print("  Name:", wio_site.pop_item("name", None))
    _print("  Contents:", [k for k in wio_site.keys()])
    _print("  Ignoring boundaries", level=2)

    # read energy_resource:
    energy_resource = wio_site["energy_resource"]
    _print("  Reading energy_resource", level=2)
    _print("    Name:", energy_resource.pop_item("name", None), level=2)
    _print("    Contents:", [k for k in energy_resource.keys()], level=2)

    # read wind_resource:
    wind_resource = energy_resource["wind_resource"]
    _print("    Reading wind_resource", level=3)
    _print("      Name:", wind_resource.pop_item("name", None), level=3)
    _print("      Contents:", [k for k in wind_resource.keys()], level=3)

    # read fields
    coords = Dict(_name="coords")
    fields = Dict(_name="fields")
    dims = Dict(_name="dims")
    for n, d in wind_resource.items():
        read_wind_resource_field(n, d, coords, fields, dims, verbosity)

    # special case: operating field
    if FV.OPERATING in fields:
        wio_dict["wind_farm"][FV.OPERATING] = (
            dims.pop(FV.OPERATING),
            fields.pop(FV.OPERATING),
        )
        if FC.TURBINE in coords:
            if not any([FC.TURBINE in dms for dms in dims.values()]):
                if verbosity > 2:
                    print(
                        f"      Removing coordinate '{FC.TURBINE}', since only relevant for operating flag"
                    )
                coords.pop(FC.TURBINE)

    if verbosity > 2:
        print("      Coords:")
        for c, d in coords.items():
            print(f"        {c}: Shape {d.shape}")
        print("      Fields:")
        for f, d in dims.items():
            if len(d):
                print(f"        {f}: Dims {d}, shape {fields[f].shape}")
            else:
                print(f"        {f} = {fields[f]}")

    return get_states(coords, fields, dims, verbosity)



"""
Model collection.
"""

from . import turbine_types as turbine_types
from . import rotor_models as rotor_models
from . import turbine_models as turbine_models
from . import farm_models as farm_models
from . import partial_wakes as partial_wakes
from . import wake_frames as wake_frames
from . import wake_models as wake_models
from . import wake_deflections as wake_deflections
from . import wake_superpositions as wake_superpositions
from . import farm_controllers as farm_controllers
from . import vertical_profiles as vertical_profiles
from . import point_models as point_models
from . import axial_induction as axial_induction
from . import ground_models as ground_models

from .model_book import ModelBook as ModelBook



import numpy as np


def weibull_weights(ws, ws_deltas, A, k):
    """
    Computes the weibull weights for given wind speeds

    Parameters
    ----------
    ws: numpy.ndarray
        The wind speed bin centre values
    ws_deltas: numpy.ndarray
        The wind speed bin widths, same shape as ws
    A: numpy.ndarray
        The Weibull scale parameters, same shape as ws
    k: numpy.ndarray
        The Weibull shape parameters, same shape as ws

    Returns
    -------
    weights: numpy.ndarray
        The weights, same shape as ws

    :group: utils

    """
    wsA = ws / A
    return ws_deltas * (k / A * wsA ** (k - 1) * np.exp(-(wsA**k)))



"""
Abstract classes and core functionality.
"""

from .model import Model as Model
from .data_calc_model import DataCalcModel as DataCalcModel
from .wind_farm import WindFarm as WindFarm
from .algorithm import Algorithm as Algorithm
from .rotor_model import RotorModel as RotorModel
from .farm_model import FarmModel as FarmModel
from .turbine_model import TurbineModel as TurbineModel
from .turbine_type import TurbineType as TurbineType
from .farm_controller import FarmController as FarmController
from .turbine import Turbine as Turbine
from .partial_wakes_model import PartialWakesModel as PartialWakesModel
from .wake_frame import WakeFrame as WakeFrame
from .wake_deflection import WakeDeflection as WakeDeflection
from .vertical_profile import VerticalProfile as VerticalProfile
from .axial_induction_model import AxialInductionModel as AxialInductionModel
from .ground_model import GroundModel as GroundModel

from .data import Data as Data
from .data import MData as MData
from .data import FData as FData
from .data import TData as TData

from .engine import Engine as Engine
from .engine import get_engine as get_engine
from .engine import has_engine as has_engine
from .engine import reset_engine as reset_engine

from .states import States as States
from .states import ExtendedStates as ExtendedStates

from .farm_data_model import FarmDataModel as FarmDataModel
from .farm_data_model import FarmDataModelList as FarmDataModelList

from .point_data_model import PointDataModel as PointDataModel
from .point_data_model import PointDataModelList as PointDataModelList

from .wake_model import WakeModel as WakeModel
from .wake_model import SingleTurbineWakeModel as SingleTurbineWakeModel
from .wake_model import TurbineInductionModel as TurbineInductionModel
from .wake_model import WakeK as WakeK

from .wake_superposition import WakeSuperposition as WakeSuperposition
from .wake_superposition import (
    WindVectorWakeSuperposition as WindVectorWakeSuperposition,
)



import pandas as pd
from pathlib import Path


class ReaderWRG:
    """
    A reader for WRG files

    Attributes
    ----------
    fpath: pathlib.Path
        Path to the wrg file

    """

    def __init__(self, fpath):
        """
        Constructor

        Parameters
        ----------
        fpath: str
            Path to the wrg file
        """
        self.fpath = Path(fpath)
        self._prepare()

    def _prepare(self):
        # read the first two lines
        with open(self.fpath, "r") as fstream:
            self._nx, self._ny, self._utmx0, self._utmy0, self._res = (
                fstream.readline().split()
            )
            second_line = fstream.readline().split()
            n_cols = len(second_line)
            self._n_sectors = int(second_line[8])
            self._nx, self._ny, self._utmx0, self._utmy0, self._res = (
                int(self._nx),
                int(self._ny),
                float(self._utmx0),
                float(self._utmy0),
                int(self._res),
            )

        def cols_sel(name, secs):
            """little helper to create column names"""
            return [f"{name}_{i}" for i in range(secs)]

        cols = [0] * (8 + 3 * self._n_sectors)
        cols[0] = "utmx"
        cols[1] = "utmy"
        cols[2] = "z"
        cols[3] = "h"
        cols[4] = "A"
        cols[5] = "K"
        cols[6] = "pw"
        cols[7] = "n_sectors"
        cols[8::3] = cols_sel("fs", self._n_sectors)
        cols[9::3] = cols_sel("As", self._n_sectors)
        cols[10::3] = cols_sel("Ks", self._n_sectors)

        self._data = pd.read_csv(
            self.fpath, names=cols, skiprows=1, sep=r"\s+", usecols=range(1, n_cols)
        )

        self._data[cols_sel("fs", self._n_sectors)] /= 10
        self._data[cols_sel("As", self._n_sectors)] /= 10
        self._data[cols_sel("Ks", self._n_sectors)] /= 100

        if len(self._data.index) != self._nx * self._ny:
            raise ValueError(
                f"Expecting {self._nx * self._ny} rows in data, got {len(self._data.index)}"
            )

    @property
    def data(self):
        """
        The WRG data

        Returns
        -------
        df: pandas.DataFrame
            The WRG data

        """
        return self._data

    @property
    def nx(self):
        """
        The number of points in x direction

        Returns
        -------
        n: int
            The number of points in x direction

        """
        return self._nx

    @property
    def ny(self):
        """
        The number of points in y direction

        Returns
        -------
        n: int
            The number of points in y direction

        """
        return self._ny

    @property
    def x0(self):
        """
        The lower left x coordinate

        Returns
        -------
        x: float
            The lower left x coordinate

        """
        return self._utmx0

    @property
    def y0(self):
        """
        The lower left y coordinate

        Returns
        -------
        y: float
            The lower left y coordinate

        """
        return self._utmy0

    @property
    def n_sectors(self):
        """
        The number of wind direction sectors

        Returns
        -------
        n: int
            The number of wind direction sectors

        """
        return self._n_sectors

    @property
    def resolution(self):
        """
        The horizontal resolution

        Returns
        -------
        res: float
            The horizontal resolution

        """
        return self._res



from pathlib import Path
import inspect
import argparse
import os

from foxes.utils import load_module


def test():
    thisdir = Path(inspect.getabsfile(inspect.currentframe())).parent
    print("TESTDIR:", thisdir)

    rdir = thisdir.parent.parent / "examples"
    rpath = rdir / "run_all.py"
    print(rpath)

    if rpath.is_file():
        run_all = load_module("run_all", rpath)

        args = argparse.Namespace()
        args.include = None
        args.exclude = ["windio"]
        args.incopt = False
        args.forceopt = False
        args.step = 0
        args.dry = False
        args.Dry = False
        args.nofig = True

        os.chdir(rdir)
        run_all.run(args)


if __name__ == "__main__":
    test()



from xarray import Dataset

from foxes.core import Engine
import foxes.constants as FC

from .pool import _write_chunk_results


class SingleChunkEngine(Engine):
    """
    Runs computations in a single chunk.

    :group: engines

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Additional parameters for the base class
        kwargs: dict, optional
            Additional parameters for the base class

        """
        ignr = ["chunk_size_states", "chunk_size_points", "n_procs"]
        for k in ignr:
            if kwargs.pop(k, None) is not None and kwargs.get("verbosity", 1) > 1:
                print(f"{type(self).__name__}: Ignoring {k}")
        super().__init__(
            *args,
            chunk_size_states=None,
            chunk_size_points=None,
            n_procs=1,
            **kwargs,
        )
        self.progress_bar = None

    def __repr__(self):
        return f"{type(self).__name__}()"

    def submit(self, f, *args, **kwargs):
        """
        Submits a job to worker, obtaining a future

        Parameters
        ----------
        f: Callable
            The function f(*args, **kwargs) to be
            submitted
        args: tuple, optional
            Arguments for the function
        kwargs: dict, optional
            Arguments for the function

        Returns
        -------
        future: object
            The future object

        """
        return f(*args, **kwargs)

    def await_result(self, future):
        """
        Waits for result from a future

        Parameters
        ----------
        future: object
            The future

        Returns
        -------
        result: object
            The calculation result

        """
        return future

    def map(
        self,
        func,
        inputs,
        *args,
        **kwargs,
    ):
        """
        Runs a function on a list of files

        Parameters
        ----------
        func: Callable
            Function to be called on each file,
            func(input, *args, **kwargs) -> data
        inputs: array-like
            The input data list
        args: tuple, optional
            Arguments for func
        kwargs: dict, optional
            Keyword arguments for func

        Returns
        -------
        results: list
            The list of results

        """
        return [func(input, *args, **kwargs) for input in inputs]

    def run_calculation(
        self,
        algo,
        model,
        model_data,
        farm_data=None,
        point_data=None,
        out_vars=[],
        chunk_store={},
        sel=None,
        isel=None,
        iterative=False,
        write_nc=None,
        **calc_pars,
    ):
        """
        Runs the model calculation

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm object
        model: foxes.core.DataCalcModel
            The model that whose calculate function
            should be run
        model_data: xarray.Dataset
            The initial model data
        farm_data: xarray.Dataset, optional
            The initial farm data
        point_data: xarray.Dataset, optional
            The initial point data
        out_vars: list of str, optional
            Names of the output variables
        chunk_store: foxes.utils.Dict
            The chunk store
        sel: dict, optional
            Selection of coordinate subsets
        isel: dict, optional
            Selection of coordinate subsets index values
        iterative: bool
            Flag for use within the iterative algorithm
        write_nc: dict, optional
            Parameters for writing results to netCDF files, e.g.
            {'out_dir': 'results', 'base_name': 'calc_results',
            'ret_data': False, 'split': 1000}.

            The split parameter controls how the output is split:
            - 'chunks': one file per chunk (fastest method),
            - 'input': split according to sizes of multiple states input files,
            - int: split with this many states per file,
            - None: create a single output file.

            Use ret_data = False together with non-single file writing
            to avoid constructing the full Dataset in memory.
        calc_pars: dict, optional
            Additional parameters for the model.calculate()

        Returns
        -------
        results: xarray.Dataset
            The model results

        """
        # subset selection:
        model_data, farm_data, point_data = self.select_subsets(
            model_data, farm_data, point_data, sel=sel, isel=isel
        )

        # basic checks:
        super().run_calculation(algo, model, model_data, farm_data, point_data)

        # prepare:
        n_states = model_data.sizes[FC.STATE]
        n_targets = point_data.sizes[FC.TARGET] if point_data is not None else 0
        out_coords = model.output_coords()
        coords = {}
        if FC.STATE in out_coords and FC.STATE in model_data.coords:
            coords[FC.STATE] = model_data[FC.STATE].to_numpy()
        algo.reset_chunk_store(chunk_store)

        # calculate:

        if n_states > 1:
            self.print(
                f"{type(self).__name__}: Running single chunk calculation for {n_states} states"
            )

        data = self.get_chunk_input_data(
            algo=algo,
            model_data=model_data,
            farm_data=farm_data,
            point_data=point_data,
            states_i0_i1=(0, n_states),
            targets_i0_i1=(0, n_targets),
            out_vars=out_vars,
            chunki_states=0,
            chunki_points=0,
            n_chunks_states=1,
            n_chunks_points=0,
        )

        results = model.calculate(algo, *data, **calc_pars)
        results = _write_chunk_results(algo, results, write_nc, out_coords, data[0])
        results = {(0, 0): (results, algo.chunk_store)}
        del data

        if farm_data is None:
            farm_data = Dataset()
        goal_data = farm_data if point_data is None else point_data

        return self.combine_results(
            algo=algo,
            futures=None,
            results=results,
            model_data=model_data,
            out_vars=out_vars,
            out_coords=out_coords,
            n_chunks_states=1,
            n_chunks_targets=1,
            goal_data=goal_data,
            iterative=iterative,
            write_nc=write_nc,
        )



import pandas as pd
from pathlib import Path
import inspect

import foxes
import foxes.variables as FV


def test():
    thisdir = Path(inspect.getabsfile(inspect.currentframe())).parent
    print("TESTDIR:", thisdir)

    c = 500
    cpath = thisdir / "flappy"
    tfile = thisdir / "NREL-5MW-D126-H90.csv"
    sfile = thisdir / "states.csv.gz"
    lfile = thisdir / "test_farm.csv"
    cases = [
        (["Basta"], "centre", "rotor_points"),
        (["Basta"], "grid4", "grid4"),
        (["Basta"], "grid16", "grid16"),
        (["Basta"], "grid64", "grid64"),
    ]

    with foxes.Engine.new("threads", chunk_size_states=c):
        for i, (wakes, rotor, pwake) in enumerate(cases):
            print(f"\nENTERING CASE {(wakes, rotor, pwake)}\n")

            mbook = foxes.models.ModelBook()
            ttype = foxes.models.turbine_types.PCtFile(
                data_source=tfile, var_ws_ct=FV.REWS, var_ws_P=FV.REWS
            )
            mbook.turbine_types[ttype.name] = ttype

            mbook.wake_models["Basta"] = foxes.models.wake_models.wind.Bastankhah2014(
                sbeta_factor=0.25, superposition="ws_linear", induction="Betz"
            )

            states = foxes.input.states.StatesTable(
                data_source=sfile,
                output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
                var2col={FV.WS: "ws", FV.WD: "wd", FV.TI: "ti"},
                fixed_vars={FV.RHO: 1.225, FV.Z0: 0.1, FV.H: 100.0},
                profiles={FV.WS: "ABLLogNeutralWsProfile"},
            )

            farm = foxes.WindFarm()
            foxes.input.farm_layout.add_from_file(
                farm,
                lfile,
                turbine_models=["kTI_amb_02", ttype.name],
                verbosity=1,
            )

            algo = foxes.algorithms.Downwind(
                farm,
                states,
                mbook=mbook,
                rotor_model=rotor,
                wake_models=wakes,
                wake_frame="rotor_wd",
                partial_wakes=pwake,
                verbosity=1,
            )

            data = algo.calc_farm()

            df = data.to_dataframe()[
                [FV.AMB_WD, FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P]
            ]

            cfile = cpath / f"results_{i}.csv.gz"
            print("\nReading file", cfile)
            fdata = pd.read_csv(cfile).set_index(["state", "turbine"])

            print()
            print("TRESULTS\n")
            print(df)
            print(fdata)

            print("\nVERIFYING\n")
            df[FV.WS] = df["REWS"]
            df[FV.AMB_WS] = df["AMB_REWS"]

            delta = df - fdata
            print(delta)
            chk = delta[[FV.AMB_WS, FV.AMB_P, FV.WS, FV.P]]
            print(chk)
            chk = chk.abs()
            print(chk.max())

            var = FV.AMB_WS
            sel = chk[var] >= 1e-7
            print(f"\nCHECKING {var}, {(wakes, rotor, pwake)}\n")
            print(df.loc[sel])
            print(fdata.loc[sel])
            print(delta.loc[sel])
            assert (chk[var] < 1e-7).all()

            var = FV.AMB_P
            sel = chk[var] >= 1e-5
            print(f"\nCHECKING {var}, {(wakes, rotor, pwake)}\n")
            print(df.loc[sel])
            print(fdata.loc[sel])
            print(delta.loc[sel])
            assert (chk[var] < 1e-5).all()

            var = FV.WS
            sel = chk[var] >= 1.7e-3
            print(f"\nCHECKING {var}, {(wakes, rotor, pwake)}\n")
            print(df.loc[sel])
            print(fdata.loc[sel])
            print(delta.loc[sel])
            assert (chk[var] < 1.7e-3).all()

            var = FV.P
            sel = chk[var] >= 1.51
            print(f"\nCHECKING {var}, {(wakes, rotor, pwake)}\n")
            print(df.loc[sel])
            print(fdata.loc[sel])
            print(delta.loc[sel])
            assert (chk[var] < 1.51).all()

            print()


if __name__ == "__main__":
    test()



import numpy as np

from foxes.output import SliceData
import foxes.variables as FV

from .get_fig import get_fig


class FlowPlots2D(SliceData):
    """
    Class for horizontal or vertical 2D flow plots

    :group: output.flow_plots_2d

    """

    def get_mean_fig_xy(
        self,
        var,
        xlabel="x [m]",
        ylabel="y [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        ret_data=False,
        animated=False,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a horizontal xy-plane.

        Parameters
        ----------
        var: str
            The variable name
        xlabel: str, optional
            The x axis label
        ylabel: str, optional
            The y axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at each `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        ret_data: bool
            Flag for returning image data
        animated: bool
            Switch for usage for an animation
        kwargs: dict, optional
            Additional parameters for SliceData.get_mean_data_xy

        Returns
        -------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet
        data: numpy.ndarray, optional
            The image data, shape: (n_x, n_y)

        """
        if self.nofig:
            return None

        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        data, gdata = self.get_mean_data_xy(
            variables=variables,
            vmin={var: vmin} if vmin is not None else {},
            vmax={var: vmax} if vmax is not None else {},
            data_format="numpy",
            ret_grid=True,
            **kwargs,
        )
        x_pos, y_pos, z_pos, __ = gdata

        if title is None:
            title = f"States mean, z =  {int(np.round(z_pos))} m"

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[None, :, :, wdi],
                data[None, :, :, wsi],
            )
        )

        # create plot:
        out = get_fig(
            var=var,
            fig=fig,
            figsize=figsize,
            ax=ax,
            data=data[None, :, :, vi],
            si=0,
            s=None,
            levels=levels,
            x_pos=x_pos,
            y_pos=y_pos,
            cmap=cmap,
            xlabel=xlabel,
            ylabel=ylabel,
            title=title,
            add_bar=add_bar,
            vlabel=vlabel,
            quiv=quiv,
            vmin=vmin,
            vmax=vmax,
            ret_state=ret_state,
            ret_im=ret_im,
            animated=animated,
        )

        if ret_data:
            out = list(out) if isinstance(out, tuple) else [out]
            return tuple(out + [data[:, :, 0]])

        return out

    def get_mean_fig_xz(
        self,
        var,
        x_direction=270,
        xlabel="x [m]",
        zlabel="z [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        ret_data=False,
        animated=False,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a horizontal xz-plane.

        Parameters
        ----------
        var: str
            The variable name
        x_direction: float, optional
            The direction of the x axis, 0 = north
        xlabel: str, optional
            The x axis label
        zlabel: str, optional
            The z axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at each `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        ret_data: bool
            Flag for returning image data
        animated: bool
            Switch for usage for an animation
        kwargs: dict, optional
            Additional parameters for SliceData.get_mean_data_xz

        Returns
        -------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet
        data: numpy.ndarray, optional
            The image data, shape: (n_x, n_y)

        """
        if self.nofig:
            return None

        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        data, gdata = self.get_mean_data_xz(
            variables=variables,
            vmin={var: vmin} if vmin is not None else {},
            vmax={var: vmax} if vmax is not None else {},
            x_direction=x_direction,
            data_format="numpy",
            ret_grid=True,
            **kwargs,
        )
        x_pos, y_pos, z_pos, __ = gdata

        if title is None:
            title = f"States mean, x direction {x_direction}°, y =  {int(np.round(y_pos))} m"

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[None, :, :, wdi],
                data[None, :, :, wsi],
            )
        )

        # create plot:
        out = get_fig(
            var=var,
            fig=fig,
            figsize=figsize,
            ax=ax,
            data=data[None, :, :, vi],
            si=0,
            s=None,
            levels=levels,
            x_pos=x_pos,
            y_pos=z_pos,
            cmap=cmap,
            xlabel=xlabel,
            ylabel=zlabel,
            title=title,
            add_bar=add_bar,
            vlabel=vlabel,
            vmin=vmin,
            vmax=vmax,
            ret_state=ret_state,
            ret_im=ret_im,
            quiv=quiv,
            animated=animated,
        )

        if ret_data:
            out = list(out) if isinstance(out, tuple) else [out]
            return tuple(out + [data[:, :, 0]])

        return out

    def get_mean_fig_yz(
        self,
        var,
        x_direction=270,
        ylabel="x [m]",
        zlabel="z [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        ret_data=False,
        animated=False,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a horizontal yz-plane.

        Parameters
        ----------
        var: str
            The variable name
        x_direction: float, optional
            The direction of the x axis, 0 = north
        ylabel: str, optional
            The y axis label
        zlabel: str, optional
            The z axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at each `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        ret_data: bool
            Flag for returning image data
        animated: bool
            Switch for usage for an animation
        kwargs: dict, optional
            Additional parameters for SliceData.get_mean_data_yz

        Returns
        -------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet
        data: numpy.ndarray, optional
            The image data, shape: (n_x, n_y)

        """
        if self.nofig:
            return None

        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        data, gdata = self.get_mean_data_yz(
            variables=variables,
            vmin={var: vmin} if vmin is not None else {},
            vmax={var: vmax} if vmax is not None else {},
            x_direction=x_direction,
            data_format="numpy",
            ret_grid=True,
            **kwargs,
        )
        x_pos, y_pos, z_pos, __ = gdata

        if title is None:
            title = f"States mean, x direction {x_direction}°, x =  {int(np.round(x_pos))} m"

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[None, :, :, wdi],
                data[None, :, :, wsi],
            )
        )

        # create plot:
        out = get_fig(
            var=var,
            fig=fig,
            figsize=figsize,
            ax=ax,
            data=data[None, :, :, vi],
            si=0,
            s=None,
            levels=levels,
            x_pos=y_pos,
            y_pos=z_pos,
            cmap=cmap,
            xlabel=ylabel,
            ylabel=zlabel,
            title=title,
            add_bar=add_bar,
            vlabel=vlabel,
            vmin=vmin,
            vmax=vmax,
            ret_state=ret_state,
            ret_im=ret_im,
            quiv=quiv,
            invert_axis="x",
            animated=animated,
        )

        if ret_data:
            out = list(out) if isinstance(out, tuple) else [out]
            return tuple(out + [data[:, :, 0]])

        return out

    def gen_states_fig_xy(
        self,
        var,
        xlabel="x [m]",
        ylabel="y [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        animated=False,
        rotor_color=None,
        precalc=False,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a horizontal xy-plane.

        Parameters
        ----------
        var: str
            The variable name
        xlabel: str, optional
            The x axis label
        ylabel: str, optional
            The y axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at each `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        animated: bool
            Switch for usage for an animation
        rotor_color: str, optional
            Indicate the rotor orientation by a colored line
        precalc: bool or tuple
            Flag for pre-calculation run, adding an additional
            generator call before the actual plot generations,
            yields data, states, gdata. The same tuple can be given
            for avoiding its calculation and picking up from there.
        kwargs: dict, optional
            Additional parameters for SliceData.get_states_data_xy

        Yields
        ------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet

        """
        if self.nofig:
            yield None

        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        if isinstance(precalc, (tuple, list)):
            data, states, gdata = precalc
        else:
            data, states, gdata = self.get_states_data_xy(
                variables=variables,
                vmin={var: vmin} if vmin is not None else {},
                vmax={var: vmax} if vmax is not None else {},
                data_format="numpy",
                ret_states=True,
                ret_grid=True,
                **kwargs,
            )
            if precalc:
                yield data, states, gdata

        x_pos, y_pos, z_pos, __ = gdata

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[..., wdi],
                data[..., wsi],
            )
        )

        # loop over states:
        for si, s in enumerate(states):
            if animated and si == 0:
                vmin = vmin if vmin is not None else np.min(data[..., vi])
                vmax = vmax if vmax is not None else np.max(data[..., vi])
            elif animated and si > 0:
                add_bar = False

            if not animated and title is None:
                ttl = f"State {s}"
                ttl += f", z =  {int(np.round(z_pos))} m"
            elif callable(title):
                ttl = title(si, s)
            else:
                ttl = title

            # get data for show_turbines
            if rotor_color is not None:
                try:
                    turb_angle = self.fres[FV.AMB_WD][si] + self.fres[FV.YAWM][si]
                except KeyError:
                    turb_angle = self.fres[FV.AMB_WD][si]

                show_rotor_dict = {
                    "color": rotor_color,
                    "D": self.fres[FV.D][si],
                    "H": self.fres[FV.H][si],
                    "X": self.fres[FV.X][si],
                    "Y": self.fres[FV.Y][si],
                    "AMB_WD": self.fres[FV.AMB_WD][si],
                    "turb_angle": turb_angle,
                }
            else:
                show_rotor_dict = None

            out = get_fig(
                var=var,
                fig=fig,
                figsize=figsize,
                ax=ax,
                data=data[..., vi],
                si=si,
                s=s,
                levels=levels,
                x_pos=x_pos,
                y_pos=y_pos,
                cmap=cmap,
                xlabel=xlabel,
                ylabel=ylabel,
                title=ttl,
                add_bar=add_bar,
                vlabel=vlabel,
                vmin=vmin,
                vmax=vmax,
                quiv=quiv,
                ret_state=ret_state,
                ret_im=ret_im,
                animated=animated,
                show_rotor_dict=show_rotor_dict,
            )

            yield out

    def gen_states_fig_xz(
        self,
        var,
        x_direction=270.0,
        xlabel="x [m]",
        zlabel="z [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        animated=False,
        rotor_color=None,
        precalc=False,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a vertical xz-plane.

        Parameters
        ----------
        var: str
            The variable name
        x_direction: float, optional
            The direction of the x axis, 0 = north
        xlabel: str, optional
            The x axis label
        zlabel: str, optional
            The z axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at ech `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        animated: bool
            Switch for usage for an animation
        rotor_color: str, optional
            Indicate the rotor orientation by a colored line
        precalc: bool or tuple
            Flag for pre-calculation run, adding an additional
            generator call before the actual plot generations,
            yields data, states, gdata. The same tuple can be given
            for avoiding its calculation and picking up from there.
        kwargs: dict, optional
            Additional parameters for SliceData.get_states_data_xz

        Yields
        ------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet

        """
        if self.nofig:
            yield None

        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        if isinstance(precalc, (tuple, list)):
            data, states, gdata = precalc
        else:
            data, states, gdata = self.get_states_data_xz(
                variables=variables,
                vmin={var: vmin} if vmin is not None else {},
                vmax={var: vmax} if vmax is not None else {},
                data_format="numpy",
                ret_states=True,
                ret_grid=True,
                x_direction=x_direction,
                **kwargs,
            )
            if precalc:
                yield data, states, gdata

        x_pos, y_pos, z_pos, __ = gdata

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[..., wdi],
                data[..., wsi],
            )
        )

        # loop over states:
        for si, s in enumerate(states):
            if animated and si > 0 and vmin is not None and vmax is not None:
                add_bar = False
            if not animated and title is None:
                ttl = f"State {s}"
                ttl += f", x direction = {x_direction}°"
                ttl += f", y =  {int(np.round(y_pos))} m"
            elif callable(title):
                ttl = title(si, s)
            else:
                ttl = title

            # get data for show_turbines
            if rotor_color is not None:
                try:
                    turb_angle = self.fres[FV.AMB_WD][si] + self.fres[FV.YAWM][si]
                except KeyError:
                    turb_angle = self.fres[FV.AMB_WD][si]

                show_rotor_dict = {
                    "color": rotor_color,
                    "D": self.fres[FV.D][si],
                    "H": self.fres[FV.H][si],
                    "X": self.fres[FV.X][si],
                    "Y": self.fres[FV.Y][si],
                    "AMB_WD": self.fres[FV.AMB_WD][si],
                    "turb_angle": turb_angle,
                }
            else:
                show_rotor_dict = None

            out = get_fig(
                var=var,
                fig=fig,
                figsize=figsize,
                ax=ax,
                data=data[..., vi],
                si=si,
                s=s,
                levels=levels,
                x_pos=x_pos,
                y_pos=z_pos,
                cmap=cmap,
                xlabel=xlabel,
                ylabel=zlabel,
                title=ttl,
                add_bar=add_bar,
                vlabel=vlabel,
                quiv=quiv,
                vmin=vmin,
                vmax=vmax,
                ret_state=ret_state,
                ret_im=ret_im,
                animated=animated,
                show_rotor_dict=show_rotor_dict,
            )

            yield out

    def gen_states_fig_yz(
        self,
        var,
        x_direction=270.0,
        ylabel="y [m]",
        zlabel="z [m]",
        levels=None,
        figsize=None,
        title=None,
        vlabel=None,
        fig=None,
        ax=None,
        add_bar=True,
        cmap=None,
        vmin=None,
        vmax=None,
        quiver_n=None,
        quiver_pars={},
        ret_state=False,
        ret_im=False,
        animated=False,
        rotor_color=None,
        precalc=False,
        **kwargs,
    ):
        """
        Generates 2D farm flow figure in a vertical yz-plane.

        Parameters
        ----------
        var: str
            The variable name
        x_direction: float, optional
            The direction of the x axis, 0 = north
        ylabel: str, optional
            The y axis label
        zlabel: str, optional
            The z axis label
        levels: int, optional
            The number of levels for the contourf plot, or None for pure image
        figsize: tuple, optional
            The figsize for plt.Figure
        title: str, optional
            The title
        vlabel: str, optional
            The variable label
        fig: plt.Figure, optional
            The figure object
        ax: plt.Axes, optional
            The figure axes
        add_bar: bool
            Add a color bar
        cmap: str, optional
            The colormap
        vmin: float, optional
            The minimal variable value
        vmax: float, optional
            The maximal variable value
        quiver_n: int, optional
            Place a vector at ech `n`th point
        quiver_pars: dict, optional
            Parameters for plt.quiver
        ret_state: bool
            Flag for state index return
        ret_im: bool
            Flag for image return
        animated: bool
            Switch for usage for an animation
        rotor_color: str, optional
            Indicate the rotor orientation by a colored line
        precalc: bool or tuple
            Flag for pre-calculation run, adding an additional
            generator call before the actual plot generations,
            yields data, states, gdata. The same tuple can be given
            for avoiding its calculation and picking up from there.
        kwargs: dict, optional
            Additional parameters for SliceData.get_states_data_yz

        Yields
        ------
        fig: matplotlib.Figure
            The figure object
        si: int, optional
            The state index
        im: tuple, optional
            The image objects, matplotlib.collections.QuadMesh
            or matplotlib.QuadContourSet

        """
        if self.nofig:
            yield None

        variables = list(set([var] + [FV.WD, FV.WS]))
        vi = variables.index(var)
        wdi = variables.index(FV.WD)
        wsi = variables.index(FV.WS)

        if isinstance(precalc, (tuple, list)):
            data, states, gdata = precalc
        else:
            data, states, gdata = self.get_states_data_yz(
                variables=variables,
                vmin={var: vmin} if vmin is not None else {},
                vmax={var: vmax} if vmax is not None else {},
                data_format="numpy",
                ret_states=True,
                ret_grid=True,
                x_direction=x_direction,
                **kwargs,
            )
            if precalc:
                yield data, states, gdata

        x_pos, y_pos, z_pos, __ = gdata

        # define wind vector arrows:
        qpars = dict(angles="xy", scale_units="xy", scale=0.05)
        qpars.update(quiver_pars)
        quiv = (
            None
            if quiver_n is None
            else (
                quiver_n,
                qpars,
                data[..., wdi],
                data[..., wsi],
            )
        )

        # loop over states:
        for si, s in enumerate(states):
            if animated and si > 0 and vmin is not None and vmax is not None:
                add_bar = False
            if not animated and title is None:
                ttl = f"State {s}" if title is None else title
                ttl += f", x direction = {x_direction}°"
                ttl += f", x =  {int(np.round(x_pos))} m"
            elif callable(title):
                ttl = title(si, s)
            else:
                ttl = title

            # get data for show_turbines
            if rotor_color is not None:
                try:
                    turb_angle = self.fres[FV.AMB_WD][si] + self.fres[FV.YAWM][si]
                except KeyError:
                    turb_angle = self.fres[FV.AMB_WD][si]

                show_rotor_dict = {
                    "color": rotor_color,
                    "D": self.fres[FV.D][si],
                    "H": self.fres[FV.H][si],
                    "X": self.fres[FV.X][si],
                    "Y": self.fres[FV.Y][si],
                    "AMB_WD": self.fres[FV.AMB_WD][si],
                    "turb_angle": turb_angle,
                }
            else:
                show_rotor_dict = None

            out = get_fig(
                var=var,
                fig=fig,
                figsize=figsize,
                ax=ax,
                data=data[..., vi],
                si=si,
                s=s,
                levels=levels,
                x_pos=y_pos,
                y_pos=z_pos,
                cmap=cmap,
                xlabel=ylabel,
                ylabel=zlabel,
                title=ttl,
                add_bar=add_bar,
                vlabel=vlabel,
                vmin=vmin,
                vmax=vmax,
                quiv=quiv,
                ret_state=ret_state,
                ret_im=ret_im,
                invert_axis="x",
                animated=animated,
                show_rotor_dict=show_rotor_dict,
            )

            yield out



import numpy as np
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    # define arguments and options:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-y", "--yawm", help="YAWM angle of first turbine", default=30.0, type=float
    )
    parser.add_argument(
        "-y2", "--yawm2", help="YAWM angle of second turbine", default=0.0, type=float
    )
    parser.add_argument("--ws", help="The wind speed", type=float, default=9.0)
    parser.add_argument("--wd", help="The wind direction", type=float, default=270.0)
    parser.add_argument("--ti", help="The TI value", type=float, default=0.08)
    parser.add_argument("--rho", help="The air density", type=float, default=1.225)
    parser.add_argument(
        "-dx", "--deltax", help="The turbine distance in x", type=float, default=1500.0
    )
    parser.add_argument(
        "-dy", "--deltay", help="Turbine layout y step", type=float, default=0.0
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["TurbOPark_vector_ambka004", "CrespoHernandez_quadratic_ambka04"],
        nargs="+",
    )
    parser.add_argument(
        "-d", "--deflection", help="The wake deflection", default="Jimenez"
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument("-f", "--frame", help="The wake frame", default="rotor_wd")
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument("-v", "--var", help="The plot variable", default=FV.WS)
    parser.add_argument(
        "-it", "--iterative", help="Use iterative algorithm", action="store_true"
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=5000,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    # create model book
    mbook = foxes.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    # set turbines in yaw
    yawm = np.array([[args.yawm, args.yawm2]])
    mbook.turbine_models["set_yawm"] = foxes.models.turbine_models.SetFarmVars(
        pre_rotor=True
    )
    mbook.turbine_models["set_yawm"].add_var(FV.YAWM, yawm)

    # create states
    states = foxes.input.states.SingleStateStates(
        ws=args.ws, wd=args.wd, ti=args.ti, rho=args.rho
    )

    # create wind farm
    print("\nCreating wind farm")
    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_row(
        farm=farm,
        xy_base=np.array([0.0, 0.0]),
        xy_step=np.array([args.deltax, args.deltay]),
        n_turbines=2,
        turbine_models=args.tmodels + ["set_yawm", "yawm2yaw", ttype.name],
    )

    # create algorithm
    Algo = foxes.algorithms.Iterative if args.iterative else foxes.algorithms.Downwind
    algo = Algo(
        farm,
        states=states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame=args.frame,
        wake_deflection=args.deflection,
        partial_wakes=args.pwakes,
        mbook=mbook,
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        # calculate farm results
        farm_results = algo.calc_farm()
        print("\nResults data:\n", farm_results)

        if not args.nofig:
            # xy horizontal flow plot
            print("\nHorizontal flow figure output:")
            o = foxes.output.FlowPlots2D(algo, farm_results)
            g = o.gen_states_fig_xy(
                args.var, resolution=10, xmin=-500, xmax=3000, rotor_color="red"
            )
            fig = next(g)
            plt.show()
            plt.close(fig)

            # yz flow plot
            print("\nVertical flow figure output:")
            o = foxes.output.FlowPlots2D(algo, farm_results)
            g = o.gen_states_fig_yz(
                args.var,
                resolution=5,
                x=750,
                ymin=-200,
                ymax=200,
                zmin=0,
                zmax=250,
                rotor_color="red",
                verbosity=0,
            )
            fig = next(g)
            plt.show()
            plt.close(fig)

        # add capacity and efficiency to farm results
        o = foxes.output.FarmResultsEval(farm_results)
        o.add_capacity(algo)
        o.add_capacity(algo, ambient=True)
        o.add_efficiency()

        # state-turbine results
        farm_df = farm_results.to_dataframe()
        print("\nFarm results data:\n")
        print(
            farm_df[
                [
                    FV.X,
                    FV.AMB_REWS,
                    FV.REWS,
                    FV.AMB_TI,
                    FV.TI,
                    FV.AMB_P,
                    FV.P,
                    FV.WD,
                    FV.YAW,
                    FV.YAWM,
                ]
            ]
        )
        print()

        # results by turbine
        turbine_results = o.reduce_states(
            {
                FV.AMB_P: "weights",
                FV.P: "weights",
                FV.AMB_CAP: "weights",
                FV.CAP: "weights",
            }
        )
        turbine_results[FV.AMB_YLD] = o.calc_turbine_yield(
            algo=algo, annual=True, ambient=True
        )
        turbine_results[FV.YLD] = o.calc_turbine_yield(algo=algo, annual=True)
        turbine_results[FV.EFF] = turbine_results[FV.P] / turbine_results[FV.AMB_P]
        print("\nResults by turbine:\n")
        print(turbine_results)

        # power results
        P0 = o.calc_mean_farm_power(ambient=True)
        P = o.calc_mean_farm_power()
        print(f"\nFarm power        : {P / 1000:.1f} MW")
        print(f"Farm ambient power: {P0 / 1000:.1f} MW")
        print(f"Farm efficiency   : {o.calc_farm_efficiency() * 100:.2f} %")
        print(f"Annual farm yield : {turbine_results[FV.YLD].sum():.2f} GWh.")



import numpy as np

from foxes.utils import Dict
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC


class Data(Dict):
    """
    Container for numpy array data and
    the associated meta data.

    Attributes
    ----------
    dims: dict
        The dimensions tuples, same or subset
        of data keys
    loop_dims: array_like of str
        List of the loop dimensions during xarray's
        `apply_ufunc` calculations
    sizes: dict
        The dimension sizes
    chunki_states: int, optional
        The index of the states chunk
    chunki_points: int, optional
        The index of the points chunk

    :group: core

    """

    def __init__(
        self,
        data={},
        dims={},
        loop_dims=[FC.STATE],
        states_i0=None,
        chunki_states=None,
        chunki_points=None,
        n_chunks_states=None,
        n_chunks_points=None,
        name="data",
    ):
        """
        Constructor.

        Parameters
        ----------
        data: dict
            The initial data to be stored
        dims: dict
            The dimensions tuples, same or subset
            of data keys
        loop_dims: array_like of str
            List of the loop dimensions during xarray's
            `apply_ufunc` calculations
        states_i0: int, optional
            The index of the first state
        chunki_states: int, optional
            The index of the states chunk
        chunki_points: int, optional
            The index of the points chunk
        n_chunks_states: int, optional
            The number of states chunks
        n_chunks_points: int, optional
            The number of points chunks
        name: str
            The data container name

        """
        super().__init__(_name=name)

        self.update(data)
        self.dims = dims
        self.loop_dims = loop_dims

        self.__states_i0 = states_i0
        self.__chunki_states = chunki_states
        self.__chunki_points = chunki_points
        self.__n_chunks_states = n_chunks_states
        self.__n_chunks_points = n_chunks_points

        self.sizes = {}
        for v, d in data.items():
            self._run_entry_checks(v, d, dims[v])

        self._auto_update()

    @property
    def n_states(self):
        """
        The number of states

        Returns
        -------
        int:
            The number of states

        """
        return self.sizes[FC.STATE] if FC.STATE in self.sizes else None

    @property
    def n_turbines(self):
        """
        The number of turbines

        Returns
        -------
        int:
            The number of turbines

        """
        return self.sizes[FC.TURBINE] if FC.TURBINE in self.sizes else None

    @property
    def chunki_states(self):
        """
        The index of the states chunk

        Returns
        -------
        int:
            The index of the states chunk

        """
        return self.__chunki_states

    @property
    def chunki_points(self):
        """
        The index of the points chunk

        Returns
        -------
        int:
            The index of the points chunk

        """
        return self.__chunki_points

    @property
    def n_chunks_states(self):
        """
        The number of states chunks

        Returns
        -------
        int:
            The number of states chunks

        """
        return self.__n_chunks_states

    @property
    def n_chunks_points(self):
        """
        The number of points chunks

        Returns
        -------
        int:
            The number of points chunks

        """
        return self.__n_chunks_points

    def states_i0(self, counter=False):
        """
        Get the state counter for first state in chunk

        Parameters
        ----------
        counter: bool
            Return the state counter instead of the index

        Returns
        -------
        int:
            The state counter for first state in chunk
            or the corresponding index

        """
        if counter:
            if self.__states_i0 is None:
                raise KeyError(f"Data '{self.name}': states_i0 requested but not set")
            return self.__states_i0
        elif FC.STATE not in self:
            return None
        else:
            return self[FC.STATE][0]

    def _auto_update(self):
        """Checks and operations after data changes"""
        data = self
        dims = self.dims

        if (
            FV.TXYH not in data
            and FV.X in data
            and FV.Y in data
            and FV.H in data
            and dims[FV.X] == dims[FV.Y]
            and dims[FV.X] == dims[FV.H]
        ):
            self[FV.TXYH] = np.stack([self[FV.X], self[FV.Y], self[FV.H]], axis=-1)

            self[FV.X] = self[FV.TXYH][..., 0]
            self[FV.Y] = self[FV.TXYH][..., 1]
            self[FV.H] = self[FV.TXYH][..., 2]

            self.dims[FV.TXYH] = tuple(list(dims[FV.X]) + [FC.XYH])

        allc = set()
        for dms in self.dims.values():
            if dms is not None:
                allc.update(dms)
        allc = allc.difference(set(data.keys()))
        for c in allc.intersection(self.sizes.keys()):
            data[c] = np.arange(self.sizes[c])
            dims[c] = (c,)

    def _run_entry_checks(self, name, data, dims):
        """Run entry checks on new data"""
        # remove axes of size 1, added by dask for extra loop dimensions:
        if dims is not None:
            if len(dims) != len(data.shape):
                for li, ld in enumerate(self.loop_dims):
                    if data.shape[li] == 1 and (len(dims) < li + 1 or dims[li] != ld):
                        self[name] = np.squeeze(data, axis=li)
            for ci, c in enumerate(dims):
                if c not in self.sizes or self.sizes[c] == 1:
                    self.sizes[c] = self[name].shape[ci]
                elif c != FC.TARGET and self[name].shape[ci] == 1:
                    pass
                elif self.sizes[c] != self[name].shape[ci]:
                    raise ValueError(
                        f"Inconsistent size for data entry '{name}', dimension '{c}': Expecting {self.sizes[c]}, found {self[name].shape[ci]} in shape {self[name].shape}"
                    )

    def add(self, name, data, dims):
        """
        Add data entry

        Parameters
        ----------
        name: str
            The data name
        data: np.ndarray
            The data
        dims: tuple of str
            The dimensions

        """
        self[name] = data
        self.dims[name] = dims
        self._run_entry_checks(name, data, dims)
        self._auto_update()

    def get_slice(self, variables, s, dim_map={}, name=None):
        """
        Get a slice of data.

        Parameters
        ----------
        variables: list of str
            The variable list that corresponds to s
        s: slice
            The slice
        dim_map: dict
            Mapping from original to new dimensions.
            If not found, same dimensions are assumed.
        name: str, optional
            The name of the data object

        Returns
        -------
        data: Data
            The new data object, containing slices

        """
        if not isinstance(variables, (list, tuple, np.ndarray)):
            variables = [variables]
        if not isinstance(s, (list, tuple, np.ndarray)):
            s = [s]

        data = {}
        dims = {}
        for v in self.keys():
            d = self.dims[v]
            if d is not None:
                hs = tuple(
                    [s[variables.index(w)] if w in variables else np.s_[:] for w in d]
                )
                data[v] = self[v][hs]
                dims[v] = (
                    tuple([dim_map.get(dd, dd) for dd in d]) if len(dim_map) else d
                )
        if name is None:
            name = self.name
        if FC.STATE in variables and self.__states_i0 is not None:
            i0 = self.states_i0(counter=True)
            a = s[variables.index(FC.STATE)]
            sts = np.arange(i0, i0 + self.n_states)[a]
            if len(sts) == 1:
                states_i0 = sts[0]
            elif np.all(sts == np.arange(sts[0], sts[0] + len(sts))):
                states_i0 = sts[0]
            else:
                raise ValueError(
                    f"Cannot determine states_i0 for states slices {a}, leading to selection {list(sts)}"
                )
        else:
            states_i0 = None

        cls = type(self)
        if issubclass(cls, Data):
            return cls(data, dims, name=name, states_i0=states_i0)
        else:
            return cls(
                data, dims, loop_dims=self.loop_dims, name=name, states_i0=states_i0
            )

    @classmethod
    def from_dataset(cls, ds, *args, callback=None, s_states=None, copy=True, **kwargs):
        """
        Create Data object from a dataset

        Parameters
        ----------
        ds: xarray.Dataset
            The dataset
        args: tuple, optional
            Additional parameters for the constructor
        callback: Function, optional
            Function f(data, dims) that manipulates
            the data and dims dicts before construction
        s_states: slice, optional
            Slice object for states
        copy: bool
            Flag for copying data
        kwargs: dict, optional
            Additional parameters for the constructor

        Returns
        -------
        data: Data
            The data object

        """
        data = {}
        dims = {}

        for c, d in ds.coords.items():
            if c == FC.STATE:
                s = np.s_[:] if s_states is None else s_states
                data[c] = d.to_numpy()[s].copy() if copy else d.to_numpy()[s]
            else:
                data[c] = d.to_numpy().copy() if copy else d.to_numpy()
            dims[c] = d.dims

        n_states = None
        for v, d in ds.data_vars.items():
            if FC.STATE in d.dims:
                if d.dims[0] != FC.STATE:
                    raise ValueError(
                        f"Expecting coordinate '{FC.STATE}' at position 0 for data variable '{v}', got {d.dims}"
                    )
                n_states = d.shape[0]
                s = np.s_[:] if s_states is None else s_states
                data[v] = d.to_numpy()[s].copy() if copy else d.to_numpy()[s]
                dims[v] = d.dims
                if v == FV.WEIGHT and d.dims == (FC.STATE,):
                    data[v] = data[v][:, None]
                    dims[v] = (FC.STATE, FC.TURBINE)
            else:
                data[v] = d.to_numpy().copy() if copy else d.to_numpy()
                dims[v] = d.dims

        if callback is not None:
            callback(data, dims)

        if FC.STATE not in data and s_states is not None and n_states is not None:
            data[FC.STATE] = np.arange(n_states)[s_states]
            dims[FC.STATE] = (FC.STATE,)

        return cls(*args, data=data, dims=dims, **kwargs)


class MData(Data):
    """
    Container for foxes model data.

    :group: core

    """

    def __init__(self, *args, name="mdata", **kwargs):
        """
        Constructor

        Parameters
        ----------
        args: tuple, optional
            Arguments for the base class
        name: str
            The data name
        kwargs: dict, optional
            Arguments for the base class

        """
        super().__init__(*args, name=name, **kwargs)


class FData(Data):
    """
    Container for foxes farm data.

    Each farm data entry has (n_states, n_turbines) shape,
    except the dimensions.

    :group: core

    """

    def __init__(self, *args, name="fdata", **kwargs):
        """
        Constructor

        Parameters
        ----------
        args: tuple, optional
            Arguments for the base class
        name: str
            The data name
        kwargs: dict, optional
            Arguments for the base class

        """
        super().__init__(*args, loop_dims=[FC.STATE], name=name, **kwargs)

    def _run_entry_checks(self, name, data, dims):
        """Run entry checks on new data"""
        super()._run_entry_checks(name, data, dims)
        data = self[name]
        dims = self.dims[name]
        if name not in self.sizes and name not in [FC.TNAME, FV.WEIGHT]:
            dms = (FC.STATE, FC.TURBINE)
            shp = (self.n_states, self.n_turbines)
            if len(data.shape) < 2:
                raise ValueError(
                    f"FData '{self.name}': Invalid shape for '{name}', expecting {shp}, got {data.shape}"
                )
            if len(dims) < 2 or dims[:2] != dms:
                raise ValueError(
                    f"FData '{self.name}': Invalid dims for '{name}', expecting {dms}, got {dims}"
                )

    def _auto_update(self):
        """Checks and operations after data changes"""
        super()._auto_update()
        if len(self):
            for x in [FC.STATE, FC.TURBINE]:
                if x not in self.sizes:
                    raise KeyError(
                        f"FData '{self.name}': Missing '{x}' in sizes, got {sorted(list(self.sizes.keys()))}"
                    )

    @classmethod
    def from_sizes(cls, n_states, n_turbines, *args, callback=None, **kwargs):
        """
        Create Data object from model data

        Parameters
        ----------
        n_states: int
            The number of states
        n_turbines: int
            The number of turbines
        args: tuple, optional
            Additional parameters for the constructor
        callback: Function, optional
            Function f(data, dims) that manipulates
            the data and dims dicts before construction
        kwargs: dict, optional
            Additional parameters for the constructor

        Returns
        -------
        data: Data
            The data object

        """
        data = cls(*args, **kwargs)
        data.sizes[FC.STATE] = n_states
        data.sizes[FC.TURBINE] = n_turbines

        if callback is not None:
            callback(data, data.dims)

        return data

    @classmethod
    def from_mdata(cls, mdata, *args, callback=None, **kwargs):
        """
        Create Data object from model data

        Parameters
        ----------
        mdata: MData
            The model data
        args: tuple, optional
            Additional parameters for the constructor
        callback: Function, optional
            Function f(data, dims) that manipulates
            the data and dims dicts before construction
        kwargs: dict, optional
            Additional parameters for the constructor

        Returns
        -------
        data: Data
            The data object

        """
        data = cls(
            *args,
            chunki_states=mdata.chunki_states,
            chunki_points=mdata.chunki_points,
            n_chunks_states=mdata.n_chunks_states,
            n_chunks_points=mdata.n_chunks_points,
            **kwargs,
        )

        for v in [FC.STATE, FC.TURBINE]:
            data[v] = mdata[v]
            data.dims[v] = mdata.dims[v]
            data.sizes[v] = mdata.sizes[v]

        if callback is not None:
            callback(data, data.dims)

        return data

    @classmethod
    def from_dataset(cls, ds, *args, mdata=None, callback=None, **kwargs):
        """
        Create Data object from a dataset

        Parameters
        ----------
        ds: xarray.Dataset
            The dataset
        args: tuple, optional
            Additional parameters for the constructor
        mdata: MData, optional
            The mdata object
        callback: Function, optional
            Function f(data, dims) that manipulates
            the data and dims dicts before construction
        kwargs: dict, optional
            Additional parameters for the constructor

        Returns
        -------
        data: Data
            The data object

        """
        if mdata is None:
            return super().from_dataset(ds, *args, callback=callback, **kwargs)
        else:

            def cb(data, dims):
                if FC.STATE not in data:
                    data[FC.STATE] = mdata[FC.STATE]
                    dims[FC.STATE] = mdata.dims[FC.STATE]
                if FC.TURBINE not in data:
                    data[FC.TURBINE] = mdata[FC.TURBINE]
                    dims[FC.TURBINE] = mdata.dims[FC.TURBINE]
                if callback is not None:
                    callback(data, dims)

            return super().from_dataset(
                ds,
                *args,
                callback=cb,
                chunki_states=mdata.chunki_states,
                chunki_points=mdata.chunki_points,
                n_chunks_states=mdata.n_chunks_states,
                n_chunks_points=mdata.n_chunks_points,
                **kwargs,
            )


class TData(Data):
    """
    Container for foxes target data.

    Each target consists of a fixed number of
    target points.

    :group: core

    """

    def __init__(self, *args, name="tdata", **kwargs):
        """
        Constructor

        Parameters
        ----------
        args: tuple, optional
            Arguments for the base class
        name: str
            The data name
        kwargs: dict, optional
            Arguments for the base class

        """
        super().__init__(*args, loop_dims=[FC.STATE, FC.TARGET], name=name, **kwargs)

    def _run_entry_checks(self, name, data, dims):
        """Run entry checks on new data"""
        super()._run_entry_checks(name, data, dims)
        data = self[name]
        dims = self.dims[name]

        if name == FC.TARGETS:
            dms = (FC.STATE, FC.TARGET, FC.TPOINT, FC.XYH)
            shp = (self.n_states, self.n_targets, self.n_tpoints, 3)
            if dims != dms:
                raise ValueError(
                    f"TData '{self.name}': Invalid dims of {FC.TARGETS}, expecting {dms}, got {dims}"
                )
            if data.shape != shp:
                raise ValueError(
                    f"TData '{self.name}': Invalid shape of {FC.TARGETS}, expecting {shp}, got {data.shape}"
                )

        elif name == FC.TWEIGHTS:
            dms = (FC.TPOINT,)
            shp = (self.n_tpoints,)
            if dims != dms:
                raise ValueError(
                    f"TData '{self.name}': Invalid dims of {FC.TWEIGHTS}, expecting {dms}, got {dims}"
                )
            if data.shape != shp:
                raise ValueError(
                    f"TData '{self.name}': Invalid shape of {FC.TWEIGHTS}, expecting {shp}, got {data.shape}"
                )

        elif FC.TARGETS not in self:
            raise KeyError(
                f"TData '{self.name}': Missing '{FC.TARGETS}' before adding '{name}'"
            )

        elif FC.TWEIGHTS not in self:
            raise KeyError(
                f"TData '{self.name}': Missing '{FC.TWEIGHTS}' before adding '{name}'"
            )

        elif name not in self.sizes:
            dms = (FC.STATE, FC.TARGET, FC.TPOINT)
            shp = (self.n_states, self.n_targets, self.n_tpoints)
            if len(data.shape) < 3:
                raise ValueError(
                    f"TData '{self.name}': Invalid shape for '{name}', expecting {shp}, got {data.shape}"
                )
            if len(dims) < 3 or dims[:3] != dms:
                raise ValueError(
                    f"TData '{self.name}': Invalid dims for '{name}', expecting {dms}, got {dims}"
                )

    def _auto_update(self):
        """Checks and operations after data changes"""
        super()._auto_update()
        if len(self):
            for x in [FC.TARGETS, FC.TWEIGHTS]:
                if x not in self:
                    raise KeyError(
                        f"TData '{self.name}': Missing '{x}' in data, got {sorted(list(self.keys()))}"
                    )
                if x not in self.dims:
                    raise KeyError(
                        f"TData '{self.name}': Missing '{x}' in dims, got {sorted(list(self.dims.keys()))}"
                    )
            for x in [FC.STATE, FC.TARGET, FC.TPOINT]:
                if x not in self.sizes:
                    raise KeyError(
                        f"TData '{self.name}': Missing '{x}' in sizes, got {sorted(list(self.sizes.keys()))}"
                    )

    @property
    def n_targets(self):
        """
        The number of targets

        Returns
        -------
        int:
            The number of targets

        """
        return self.sizes[FC.TARGET]

    @property
    def n_tpoints(self):
        """
        The number of points per target

        Returns
        -------
        int:
            The number of points per target

        """
        return self.sizes[FC.TPOINT]

    def tpoint_mean(self, variable):
        """
        Take the mean over target points

        Parameters
        ----------
        variable: str
            The variable name

        Returns
        -------
        data: numpy.ndarray
            The reduced array, shape:
            (n_states, n_targets, ...)

        """
        return np.einsum("stp...,p->st...", self[variable], self[FC.TWEIGHTS])

    def targets_i0(self):
        """
        Get the target counter for first target in chunk

        Returns
        -------
        int:
            The target index for first target in chunk

        """
        if FC.TARGET not in self:
            return None
        else:
            return self[FC.TARGET][0]

    @classmethod
    def from_points(
        cls,
        points,
        data=None,
        dims=None,
        variables=None,
        name="tdata",
        **kwargs,
    ):
        """
        Create from points

        Parameters
        ----------
        points: np.ndarray
            The points, shape: (n_states, n_points, 3)
        data: dict, optional
            The initial data to be stored
        dims: dict, optional
            The dimensions tuples, same or subset
            of data keys
        variables: list of str
            Add default empty variables with NaN values
            and shape (n_states, n_targets, n_tpoints)
        name: str
            The data container name
        kwargs: dict, optional
            Additional parameters for the constructor

        Returns
        -------
        pdata: Data
            The data object

        """
        if len(points.shape) != 3 or points.shape[2] != 3:
            raise ValueError(
                f"Expecting points shape (n_states, n_points, 3), got {points.shape}"
            )
        data = {} if data is None else data
        dims = {} if dims is None else dims
        data[FC.TARGETS] = points[:, :, None, :]
        dims[FC.TARGETS] = (FC.STATE, FC.TARGET, FC.TPOINT, FC.XYH)
        data[FC.TWEIGHTS] = np.array([1], dtype=config.dtype_double)
        dims[FC.TWEIGHTS] = (FC.TPOINT,)
        if variables is not None:
            for v in variables:
                data[v] = np.full_like(points[:, :, None, 0], np.nan)
                dims[v] = (FC.STATE, FC.TARGET, FC.TPOINT)
        return cls(data=data, dims=dims, name=name, **kwargs)

    @classmethod
    def from_tpoints(
        cls,
        tpoints,
        tweights,
        data=None,
        dims=None,
        variables=None,
        name="tdata",
        **kwargs,
    ):
        """
        Create from points at targets

        Parameters
        ----------
        tpoints: np.ndarray
            The points at targets, shape:
            (n_states, n_targets, n_tpoints, 3)
        tweights: np.ndarray, optional
            The target point weights, shape:
            (n_tpoints,)
        data: dict, optional
            The initial data to be stored
        dims: dict, optional
            The dimensions tuples, same or subset
            of data keys
        variables: list of str
            Add default empty variables with NaN values
            and shape (n_states, n_targets, n_tpoints)
        name: str
            The data container name
        kwargs: dict, optional
            Additional parameters for the constructor

        Returns
        -------
        pdata: Data
            The data object

        """
        if len(tpoints.shape) != 4 or tpoints.shape[3] != 3:
            raise ValueError(
                f"Expecting tpoints shape (n_states, n_targets, n_tpoints, 3), got {tpoints.shape}"
            )
        data = {} if data is None else data
        dims = {} if dims is None else dims
        data[FC.TARGETS] = tpoints
        dims[FC.TARGETS] = (FC.STATE, FC.TARGET, FC.TPOINT, FC.XYH)
        data[FC.TWEIGHTS] = tweights
        dims[FC.TWEIGHTS] = (FC.TPOINT,)
        if variables is not None:
            for v in variables:
                data[v] = np.full_like(tpoints[..., 0], np.nan)
                dims[v] = (FC.STATE, FC.TARGET, FC.TPOINT)
        return cls(data=data, dims=dims, name=name, **kwargs)

    @classmethod
    def from_dataset(
        cls,
        ds,
        *args,
        s_targets=None,
        mdata=None,
        callback=None,
        **kwargs,
    ):
        """
        Create Data object from a dataset

        Parameters
        ----------
        ds: xarray.Dataset
            The dataset
        args: tuple, optional
            Additional parameters for the constructor
        s_targets: slice, optional
            Slice object for targets
        mdata: MData, optional
            The mdata object
        callback: Function, optional
            Function f(data, dims) that manipulates
            the data and dims dicts before construction
        kwargs: dict, optional
            Additional parameters for the constructor

        Returns
        -------
        data: Data
            The data object

        """
        if mdata is None:
            cb0 = callback
        else:

            def cb_mdata(data, dims):
                if FC.STATE not in data:
                    data[FC.STATE] = mdata[FC.STATE]
                    dims[FC.STATE] = mdata.dims[FC.STATE]
                if callback is not None:
                    callback(data, dims)

            cb0 = cb_mdata
            kwargs["chunki_states"] = mdata.chunki_states
            kwargs["chunki_points"] = mdata.chunki_points
            kwargs["n_chunks_states"] = mdata.n_chunks_states
            kwargs["n_chunks_points"] = mdata.n_chunks_points

        if s_targets is None:
            cb1 = cb0
        else:

            def cb_targets(data, dims):
                if FC.TARGET not in data:
                    data[FC.TARGET] = np.arange(ds.sizes[FC.TARGET])
                    dims[FC.TARGET] = (FC.TARGET,)
                for v, d in data.items():
                    if FC.TARGET in dims[v]:
                        if dims[v] == (FC.TARGET,):
                            data[v] = d[s_targets].copy()
                        elif len(dims[v]) < 3 or dims[v][:3] != (
                            FC.STATE,
                            FC.TARGET,
                            FC.TPOINT,
                        ):
                            raise ValueError(
                                f"Expecting coordinates '{(FC.STATE, FC.TARGET, FC.TPOINT)}' at positions 0-2 for data variable '{v}', got {dims[v]}"
                            )
                        else:
                            data[v] = d[:, s_targets]
                if cb0 is not None:
                    cb0(data, dims)

            cb1 = cb_targets

        return super().from_dataset(ds, *args, callback=cb1, **kwargs)



import time
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--layout",
        help="The wind farm layout file (path or static)",
        default="test_farm_67.csv",
    )
    parser.add_argument(
        "-s",
        "--states",
        help="The states input file (path or static)",
        default="wind_rose_bremen.csv",
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument(
        "-sm",
        "--sector_file",
        help="The sector management file",
        default="sector_rules.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah2014_linear_k002"],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    mbook.turbine_models["sector_rules"] = foxes.models.turbine_models.SectorManagement(
        data_source=args.sector_file,
        col_tnames="tname",
        range_vars=[FV.WD, FV.REWS],
        target_vars=[FV.MAX_P],
        colmap={"REWS_min": "WS_min", "REWS_max": "WS_max"},
    )

    states = foxes.input.states.StatesTable(
        data_source=args.states,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.WEIGHT: "weight"},
        fixed_vars={FV.RHO: 1.225, FV.TI: 0.05},
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_row(
        farm=farm,
        xy_base=[0.0, 0.0],
        xy_step=[600.0, 0.0],
        n_turbines=2,
        turbine_models=args.tmodels + [ttype.name, "sector_rules", "PMask"],
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame="rotor_wd",
        partial_wakes=args.pwakes,
        mbook=mbook,
    )

    outputs = [
        FV.D,
        FV.WD,
        FV.AMB_WD,
        FV.H,
        FV.AMB_REWS,
        FV.REWS,
        FV.AMB_P,
        FV.MAX_P,
        FV.P,
        FV.CT,
        FV.WEIGHT,
    ]

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        time0 = time.time()
        farm_results = algo.calc_farm(outputs=outputs)
        time1 = time.time()

        print("\nCalc time =", time1 - time0, "\n")

        print(farm_results)

        fr = farm_results.to_dataframe()
        sel = fr[FV.MAX_P].dropna().index
        fr = fr.loc[sel]
        print(fr)

        if not args.nofig:
            fig = plt.figure(figsize=(12, 4))
            ax1 = fig.add_subplot(121, polar=True)
            ax2 = fig.add_subplot(122, polar=True)

            o = foxes.output.RosePlotOutput(farm_results)
            o.get_figure(
                16,
                FV.P,
                [0, 100, 1000, 2000, 4000, 5001, 7000],
                turbine=0,
                title="Power turbine 0",
                fig=fig,
                ax=ax1,
            )

            o = foxes.output.RosePlotOutput(farm_results)
            o.get_figure(
                16,
                FV.P,
                [0, 100, 1000, 2000, 4000, 5001, 7000],
                turbine=1,
                title="Power turbine 1",
                fig=fig,
                ax=ax2,
            )
            plt.show()



import numpy as np
from xarray import Dataset

from foxes.core import WakeFrame, MData, FData, TData
from foxes.utils import wd2uv
from foxes.algorithms.iterative import Iterative
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC


class Timelines(WakeFrame):
    """
    Dynamic wakes for spatially uniform timeseries states.

    Attributes
    ----------
    max_length_km: float
        The maximal wake length in km
    cl_ipars: dict
        Interpolation parameters for centre line
        point interpolation
    dt_min: float
        The delta t value in minutes,
        if not from timeseries data

    :group: models.wake_frames

    """

    def __init__(self, max_length_km=2e4, cl_ipars={}, dt_min=None, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        max_length_km: float
            The maximal wake length in km
        cl_ipars: dict
            Interpolation parameters for centre line
            point interpolation
        dt_min: float, optional
            The delta t value in minutes,
            if not from timeseries data
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(max_length_km=max_length_km, **kwargs)
        self.cl_ipars = cl_ipars
        self.dt_min = dt_min

    def __repr__(self):
        return f"{type(self).__name__}(dt_min={self.dt_min}, max_length_km={self.max_length_km})"

    def _precalc_data(self, algo, states, heights, verbosity, needs_res=False):
        """Helper function for pre-calculation of ambient wind vectors"""

        if verbosity > 0:
            print(f"{self.name}: Pre-calculating ambient wind vectors")

        # get and check times:
        times = np.asarray(states.index())
        if self.dt_min is None:
            if not np.issubdtype(times.dtype, np.datetime64):
                raise TypeError(
                    f"{self.name}: Expecting state index of type np.datetime64, found {times.dtype}"
                )
            elif len(times) == 1:
                raise KeyError(
                    f"{self.name}: Expecting 'dt_min' for single step timeseries"
                )
            dt = (
                (times[1:] - times[:-1])
                .astype("timedelta64[s]")
                .astype(config.dtype_int)
            )
        else:
            n = max(len(times) - 1, 1)
            dt = np.full(n, self.dt_min * 60, dtype="timedelta64[s]").astype(
                config.dtype_int
            )

        # prepare mdata:
        data = algo.get_model_data(states)["coords"]
        mdict = {v: np.array(d) for v, d in data.items()}
        mdims = {v: (v,) for v in data.keys()}
        data = algo.get_model_data(states)["data_vars"]
        mdict.update({v: d[1] for v, d in data.items()})
        mdims.update({v: d[0] for v, d in data.items()})
        mdata = MData(mdict, mdims, loop_dims=[FC.STATE], states_i0=0)
        del mdict, mdims, data

        # prepare fdata:
        fdata = FData()

        # prepare tdata:
        n_states = states.size()
        data = {
            v: np.zeros((n_states, 1, 1), dtype=config.dtype_double)
            for v in states.output_point_vars(algo)
        }
        pdims = {v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in data.keys()}
        points = np.zeros((n_states, 1, 3), dtype=config.dtype_double)

        # calculate all heights:
        self.timelines_data = {"dxy": (("height", FC.STATE, "dir"), [])}
        weight_data = None
        for h in heights:
            if verbosity > 0:
                print(f"  Height: {h} m")

            points[..., 2] = h
            tdata = TData.from_points(
                points=points,
                data=data,
                dims=pdims,
            )

            res = states.calculate(algo, mdata, fdata, tdata)

            if weight_data is None:
                weight_data = ((FC.STATE,), tdata[FV.WEIGHT][:, 0, 0])
            elif not np.all(tdata[FV.WEIGHT] == weight_data[1]):
                raise AssertionError(
                    f"States '{self.name}': weight data mismatch between heights"
                )
            del tdata

            uv = wd2uv(res[FV.WD], res[FV.WS])[:, 0, 0, :2]
            if len(dt) == 1:
                dxy = uv * dt[0]
            else:
                dxy = uv[:-1] * dt[:, None]
                dxy = np.append(dxy, uv[-1, None, :] * dt[-1], axis=0)
            self.timelines_data["dxy"][1].append(dxy)
            """ DEBUG
            import matplotlib.pyplot as plt
            xy = np.array([np.sum(self.timelines_data[h][:n], axis=0) for n in range(len(self.timelines_data[h]))])
            print(xy)
            plt.plot(xy[:, 0], xy[:, 1])
            plt.title(f"Height {h} m")
            plt.show()
            quit()
            """

            if needs_res:
                if "U" not in self.timelines_data:
                    self.timelines_data["U"] = (("height", FC.STATE), [])
                    self.timelines_data["V"] = (("height", FC.STATE), [])
                self.timelines_data["U"][1].append(uv[:, 0])
                self.timelines_data["V"][1].append(uv[:, 1])

                for v in states.output_point_vars(algo):
                    if v not in [FV.WS, FV.WD]:
                        if v not in self.timelines_data:
                            self.timelines_data[v] = (("height", FC.STATE), [])
                        self.timelines_data[v][1].append(res[v][:, 0, 0])

            del res, uv, dxy

        self.timelines_data = Dataset(
            coords={
                FC.STATE: states.index(),
                "height": heights,
            },
            data_vars={
                v: (d[0], np.stack(d[1], axis=0))
                for v, d in self.timelines_data.items()
            },
        )

        return weight_data

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        if not isinstance(algo, Iterative):
            raise TypeError(
                f"Incompatible algorithm type {type(algo).__name__}, expecting {Iterative.__name__}"
            )
        super().initialize(algo, verbosity)

        # disable subset state selection in iterative algo:
        algo.conv_crit.disable_subsets()

        # find turbine hub heights:
        t2h = np.zeros(algo.n_turbines, dtype=config.dtype_double)
        for ti, t in enumerate(algo.farm.turbines):
            t2h[ti] = (
                t.H if t.H is not None else algo.farm_controller.turbine_types[ti].H
            )
        heights = np.unique(t2h)

        # pre-calc data:
        from foxes.input.states import OnePointFlowTimeseries

        if isinstance(algo.states, OnePointFlowTimeseries):
            self._precalc_data(algo, algo.states.base_states, heights, verbosity)
        else:
            self._precalc_data(algo, algo.states, heights, verbosity)

    def set_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to running, and moves
        all large data to stash.

        The stashed data will be returned by the
        unset_running() function after running calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Large data stash, this function adds data here, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().set_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None and (sel is not None or isel is not None):
            data_stash[self.name]["data"] = self.timelines_data

            if isel is not None:
                self.timelines_data = self.timelines_data.isel(isel)
            if sel is not None:
                self.timelines_data = self.timelines_data.sel(sel)

    def unset_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to not running, recovering large data
        from stash

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Reconstruct model data from this stash, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().unset_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None:
            data = data_stash[self.name]
            if "data" in data:
                self.timelines_data = data.pop("data")

    def calc_order(self, algo, mdata, fdata):
        """
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        order = np.zeros((fdata.n_states, fdata.n_turbines), dtype=config.dtype_int)
        order[:] = np.arange(fdata.n_turbines)[None, :]
        return order

    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        # prepare:
        targets = tdata[FC.TARGETS]
        n_states, n_targets, n_tpoints = targets.shape[:3]
        n_points = n_targets * n_tpoints
        points = targets.reshape(n_states, n_points, 3)
        rxyz = fdata[FV.TXYH][:, downwind_index]
        theights = fdata[FV.H][:, downwind_index]
        heights = self.timelines_data["height"].to_numpy()
        data_dxy = self.timelines_data["dxy"].to_numpy()

        D = np.zeros((n_states, n_points), dtype=config.dtype_double)
        D[:] = fdata[FV.D][:, downwind_index, None]

        wcoos = np.full((n_states, n_points, 3), 1e20, dtype=config.dtype_double)
        wcoosx = wcoos[:, :, 0]
        wcoosy = wcoos[:, :, 1]
        wcoos[:, :, 2] = points[:, :, 2] - rxyz[:, None, 2]

        i0 = mdata.states_i0(counter=True)
        i1 = i0 + mdata.n_states
        trace_si = np.zeros((n_states, n_points), dtype=config.dtype_int)
        trace_si[:] = i0 + np.arange(n_states)[:, None]
        for hi, h in enumerate(heights):
            dxy = data_dxy[hi][:i1]
            precond = theights[:, None] == h

            trace_p = np.zeros((n_states, n_points, 2), dtype=config.dtype_double)
            trace_p[:] = points[:, :, :2] - rxyz[:, None, :2]
            trace_l = np.zeros((n_states, n_points), dtype=config.dtype_double)
            trace_d = np.full((n_states, n_points), np.inf, dtype=config.dtype_double)
            h_trace_si = trace_si.copy()

            # flake8: noqa: F821
            def _update_wcoos(sel):
                """Local function that updates coordinates and source times"""
                nonlocal wcoosx, wcoosy, trace_si
                d = np.linalg.norm(trace_p, axis=-1)
                sel = sel & (d <= trace_d)
                if np.any(sel):
                    trace_d[sel] = d[sel]

                    nx = dxy[h_trace_si[sel]]
                    dx = np.linalg.norm(nx, axis=-1)
                    nx /= dx[:, None]
                    trp = trace_p[sel]
                    projx = np.einsum("sd,sd->s", trp, nx)

                    seln = (projx > -dx) & (projx < dx)
                    if np.any(seln):
                        wcoosx[sel] = np.where(seln, projx + trace_l[sel], wcoosx[sel])

                        ny = np.concatenate([-nx[:, 1, None], nx[:, 0, None]], axis=1)
                        projy = np.einsum("sd,sd->s", trp, ny)
                        wcoosy[sel] = np.where(seln, projy, wcoosy[sel])
                        del ny, projy

                        trace_si[sel] = np.where(seln, h_trace_si[sel], trace_si[sel])

            # step backwards in time, until wake source turbine is hit:
            _update_wcoos(precond)
            while True:
                sel = precond & (h_trace_si > 0) & (trace_l < self.max_length_km * 1e3)
                if np.any(sel):
                    h_trace_si[sel] -= 1

                    delta = dxy[h_trace_si[sel]]
                    dmag = np.linalg.norm(delta, axis=-1)
                    trace_p[sel] -= delta
                    trace_l[sel] += dmag
                    del delta, dmag

                    # check if this is closer to turbine:
                    _update_wcoos(sel)
                    del sel

                else:
                    del sel
                    break
            del trace_p, trace_l, trace_d, h_trace_si, dxy, precond

        # store turbines that cause wake:
        trace_si = np.minimum(trace_si, i0 + np.arange(n_states)[:, None])
        tdata[FC.STATE_SOURCE_ORDERI] = downwind_index

        # store states that cause wake for each target point,
        # will be used by model.get_data() during wake calculation:
        tdata.add(
            FC.STATES_SEL,
            trace_si.reshape(n_states, n_targets, n_tpoints),
            (FC.STATE, FC.TARGET, FC.TPOINT),
        )

        return algo.wake_deflection.calc_deflection(
            algo,
            mdata,
            fdata,
            tdata,
            downwind_index,
            wcoos.reshape(n_states, n_targets, n_tpoints, 3),
        )

    def get_centreline_points(self, algo, mdata, fdata, downwind_index, x):
        """
        Gets the points along the centreline for given
        values of x.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The wake frame x coordinates, shape: (n_states, n_points)

        Returns
        -------
        points: numpy.ndarray
            The centreline points, shape: (n_states, n_points, 3)

        """
        # prepare:
        n_states, n_points = x.shape
        rxyz = fdata[FV.TXYH][:, downwind_index]
        theights = fdata[FV.H][:, downwind_index]
        heights = self.timelines_data["height"].to_numpy()
        data_dxy = self.timelines_data["dxy"].to_numpy()

        points = np.zeros((n_states, n_points, 3), dtype=config.dtype_double)
        points[:] = rxyz[:, None, :]

        trace_dp = np.zeros_like(points[..., :2])
        trace_l = x.copy()
        trace_si = np.zeros((n_states, n_points), dtype=config.dtype_int)
        trace_si[:] = np.arange(n_states)[:, None]

        for hi, h in enumerate(heights):
            precond = theights == h
            if np.any(precond):
                sel = precond[:, None] & (trace_l > 0)
                while np.any(sel):
                    dxy = data_dxy[hi][trace_si[sel]]

                    trl = trace_l[sel]
                    trp = trace_dp[sel]
                    dl = np.linalg.norm(dxy, axis=-1)
                    cl = np.abs(trl - dl) < np.abs(trl)
                    if np.any(cl):
                        trace_l[sel] = np.where(cl, trl - dl, trl)
                        trace_dp[sel] = np.where(cl[:, None], trp + dxy, trp)
                    del trl, trp, dl, cl, dxy

                    trace_si[sel] -= 1
                    sel = precond[:, None] & (trace_l > 0) & (trace_si >= 0)

                si = trace_si[precond] + 1
                dxy = data_dxy[hi][si]
                dl = np.linalg.norm(dxy, axis=-1)[:, :, None]
                trl = trace_l[precond][:, :, None]
                trp = trace_dp[precond]
                sel = np.abs(trl) < 2 * dl
                trace_dp[precond] = np.where(sel, trp - trl / dl * dxy, np.nan)

                del si, dxy, dl, trl, trp, sel
            del precond
        del trace_si, trace_l

        points[..., :2] += trace_dp

        return points

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().finalize(algo, verbosity=verbosity)
        self.timelines_data = None



import numpy as np
import argparse
from pathlib import Path

import flappy as fl
from flappy.config.variables import variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c", "--chunksize", help="The maximal chunk size", type=int, default=500
    )
    parser.add_argument("-o", "--opath", help="The output file path", default=".")
    parser.add_argument(
        "--n_cpus", help="The number of processors", type=int, default=4
    )
    args = parser.parse_args()

    c = args.chunksize
    p0 = np.array([0.0, 0.0])
    stp = np.array([500.0, 0.0])
    opath = Path(args.opath)
    tfile = "../NREL-5MW-D126-H90.csv"
    sfile = "../states.csv.gz"
    lfile = "../test_farm.csv"
    cases = [
        (["Bastankhah_rotor"], ["wind_linear"], "centre"),
        (["Bastankhah_rotor"], ["wind_linear"], "grid4"),
        (["Bastankhah_rotor"], ["wind_linear"], "grid16"),
        (["Bastankhah_rotor"], ["wind_linear"], "grid64"),
    ]

    # init flappy:
    fl.init_flappy(n_cpus=args.n_cpus)

    for i, (wakes, superp, rotor) in enumerate(cases):
        print(f"\nCase {(wakes, superp, rotor)}")

        # load model book:
        mbook = fl.ModelBook(ct_power_curve_file=tfile)

        # create wind farm:
        farm = fl.WindFarm()
        fl.input.add_turbines_from_csv(
            farm,
            lfile,
            col_index="index",
            col_x="x",
            col_y="y",
            rotor_diameter=126.0,
            hub_height=90.0,
            rotor_model=rotor,
            wake_models=wakes,
            turbine_models=["ct_P_curves"],
            output_level=0,
        )

        # create states:
        states = fl.input.AFSStatesTable(
            data_file=sfile,
            col_wd="wd",
            col_ws_ref="ws",
            col_ti="ti",
            col_weight="weight",
            air_density=1.225,
            z0=0.1,
            h_ref=100.0,
            max_chunk_size=args.chunksize,
            output_level=0,
        )
        states.initialize()

        # run calculation:
        results = farm.calculate(mbook, states, wake_superp=superp, output_level=0)

        df = results.state_turbine_results[[FV.WD, FV.AMB_WS, FV.WS, FV.AMB_P, FV.P]]

        ofile = opath / f"results_{i}.csv.gz"
        print(f"Writing file {ofile}")
        df.to_csv(ofile)

    # close flappy:
    fl.shutdown_flappy()



import numpy as np
import xarray as xr
from abc import abstractmethod

from foxes.data import StaticData
from foxes.utils import Dict, new_instance
from foxes.config import config
import foxes.constants as FC

from .model import Model
from .engine import Engine


class Algorithm(Model):
    """
    Abstract base class for algorithms.

    Algorithms collect required objects for running
    calculations, and contain the calculation functions
    which are meant to be called from top level code.

    Attributes
    ----------
    verbosity: int
        The verbosity level, 0 means silent

    :group: core

    """

    def __init__(
        self,
        mbook,
        farm,
        verbosity=1,
        dbook=None,
        **engine_pars,
    ):
        """
        Constructor.

        Parameters
        ----------
        mbook: foxes.models.ModelBook
            The model book
        farm: foxes.WindFarm
            The wind farm
        verbosity: int
            The verbosity level, 0 means silent
        dbook: foxes.DataBook, optional
            The data book, or None for default
        engine_pars: dict, optional
            Parameters for the engine constructor

        """
        super().__init__()

        self.name = type(self).__name__
        self.verbosity = verbosity
        self.n_states = None
        self.n_turbines = farm.n_turbines

        self.__farm = farm
        self.__mbook = mbook
        self.__dbook = StaticData() if dbook is None else dbook
        self.__idata_mem = Dict(_name="idata_mem")
        self.__chunk_store = Dict(_name="chunk_store")

        if len(engine_pars):
            if "engine_type" in engine_pars:
                if "engine" in engine_pars:
                    raise KeyError(
                        f"{self.name}: Expecting either 'engine' or 'engine_type', not both"
                    )
            elif "engine" in engine_pars:
                engine_pars["engine_type"] = engine_pars.pop("engine")

            if "engine_type" in engine_pars:
                try:
                    e = Engine.new(verbosity=verbosity, **engine_pars)
                except TypeError as e:
                    print(f"\nError while interpreting engine_pars {engine_pars}\n")
                    raise e
                self.print(f"Algorithm '{self.name}': Selecting engine '{e}'")
                e.initialize()
            else:
                raise KeyError(
                    f"{self.name}: Found unsupported parameters {list(engine_pars.keys())}"
                )

    @property
    def farm(self):
        """
        The wind farm

        Returns
        -------
        mb: foxes.core.WindFarm
            The wind farm

        """
        return self.__farm

    @property
    def mbook(self):
        """
        The model book

        Returns
        -------
        mb: foxes.models.ModelBook()
            The model book

        """
        if self.running:
            raise ValueError(
                f"Algorithm '{self.name}': Cannot access mbook while running"
            )
        return self.__mbook

    @property
    def dbook(self):
        """
        The data book

        Returns
        -------
        mb: foxes.data.StaticData()
            The data book

        """
        if self.running:
            raise ValueError(
                f"Algorithm '{self.name}': Cannot access dbook while running"
            )
        return self.__dbook

    @property
    def idata_mem(self):
        """
        The current idata memory

        Returns
        -------
        dict :
            Keys: model name, value: idata dict

        """
        if self.running:
            raise ValueError(
                f"Algorithm '{self.name}': Cannot access idata_mem while running"
            )
        return self.__idata_mem

    @property
    def chunk_store(self):
        """
        The current chunk store

        Returns
        -------
        dict :
            Keys: model name, value: idata dict

        """
        return self.__chunk_store

    def print(self, *args, vlim=1, **kwargs):
        """
        Print function, based on verbosity.

        Parameters
        ----------
        args: tuple, optional
            Arguments for the print function
        kwargs: dict, optional
            Keyword arguments for the print function
        vlim: int
            The verbosity limit

        """
        if self.verbosity >= vlim:
            print(*args, **kwargs)

    def print_deco(self, func_name=None, n_points=None):
        """
        Helper function for printing model names

        Parameters
        ----------
        func_name: str, optional
            Name of the calling function
        n_points: int, optional
            The number of points

        """
        if self.verbosity > 0:
            deco = "-" * 60
            print(f"\n{deco}")
            print(f"  Algorithm: {type(self).__name__}")
            if func_name is not None:
                print(f"  Running {self.name}: {func_name}")
            print(deco)
            print(f"  n_states : {self.n_states}")
            print(f"  n_turbines: {self.n_turbines}")

    def initialize(self):
        """
        Initializes the algorithm.
        """
        if self.running:
            raise ValueError(
                f"Algorithm '{self.name}': Cannot initialize while running"
            )

        super().initialize(self, self.verbosity - 1)

    def store_model_data(self, model, idata, force=False):
        """
        Store model data

        Parameters
        ----------
        model: foxes.core.Model
            The model
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`
        force: bool
            Overwrite existing data

        """
        mname = f"{type(model).__name__}_{model.name}"
        if force:
            self.__idata_mem[mname] = idata
        elif mname in self.idata_mem:
            raise KeyError(f"Attempt to overwrite stored data for model '{mname}'")
        else:
            self.idata_mem[mname] = idata

    def get_model_data(self, model):
        """
        Gets model data from memory

        Parameters
        ----------
        model: foxes.core.Model
            The model

        """
        mname = f"{type(model).__name__}_{model.name}"
        try:
            return self.idata_mem[mname]
        except KeyError:
            raise KeyError(
                f"Key '{mname}' not found in idata_mem, available keys: {sorted(list(self.idata_mem.keys()))}"
            )

    def del_model_data(self, model):
        """
        Remove stored model data

        Parameters
        ----------
        model: foxes.core.Model
            The model

        """
        mname = f"{type(model).__name__}_{model.name}"
        try:
            del self.idata_mem[mname]
        except KeyError:
            raise KeyError(f"Attempt to delete data of model '{mname}', but not stored")

    def update_n_turbines(self):
        """
        Reset the number of turbines,
        according to self.farm
        """
        if self.n_turbines != self.farm.n_turbines:
            self.n_turbines = self.farm.n_turbines

            # resize stored idata, if dependent on turbine coord:
            newk = {}
            for mname, idata in self.idata_mem.items():
                if mname[:2] == "__":
                    continue
                for dname, d in idata["data_vars"].items():
                    k = f"__{mname}_{dname}_turbinv"
                    if k in self.idata_mem:
                        ok = self.idata_mem[k]
                    else:
                        ok = None
                        if FC.TURBINE in d[0]:
                            i = d[0].index(FC.TURBINE)
                            ok = np.unique(d[1], axis=1).shape[i] == 1
                        newk[k] = ok
                    if ok is not None:
                        if not ok:
                            raise ValueError(
                                f"{self.name}: Stored idata entry '{mname}:{dname}' is turbine dependent, unable to reset n_turbines"
                            )
                        if FC.TURBINE in idata["coords"]:
                            idata["coords"][FC.TURBINE] = np.arange(self.n_turbines)
                        i = d[0].index(FC.TURBINE)
                        n0 = d[1].shape[i]
                        if n0 > self.n_turbines:
                            idata["data_vars"][dname] = (
                                d[0],
                                np.take(d[1], range(self.n_turbines), axis=i),
                            )
                        elif n0 < self.n_turbines:
                            shp = [
                                d[1].shape[j] if j != i else self.n_turbines - n0
                                for j in range(len(d[1].shape))
                            ]
                            a = np.zeros(shp, dtype=d[1].dtype)
                            shp = [
                                d[1].shape[j] if j != i else 1
                                for j in range(len(d[1].shape))
                            ]
                            a[:] = np.take(d[1], -1, axis=i).reshape(shp)
                            idata["data_vars"][dname] = (
                                d[0],
                                np.append(d[1], a, axis=i),
                            )

            self.idata_mem.update(newk)

    def get_models_idata(self):
        """
        Returns idata object of models

        Returns
        -------
        idata: dict, optional
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`.
            Take algorithm's idata object by default.

        """
        if not self.initialized:
            raise ValueError(
                f"Algorithm '{self.name}': get_models_idata called before initialization"
            )
        idata = {"coords": {}, "data_vars": {}}
        for k, hidata in self.idata_mem.items():
            if len(k) < 3 or k[:2] != "__":
                idata["coords"].update(hidata["coords"])
                idata["data_vars"].update(hidata["data_vars"])
        return idata

    def get_models_data(self, idata=None, sel=None, isel=None):
        """
        Creates xarray from model input data.

        Parameters
        ----------
        idata: dict, optional
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`.
            Take algorithm's idata object by default.
        sel: dict, optional
            Selection of coordinates in dataset
        isel: dict, optional
            Selection of coordinates in dataset

        Returns
        -------
        ds: xarray.Dataset
            The model input data

        """
        if idata is None:
            idata = self.get_models_idata()
        ds = xr.Dataset(**idata)
        if isel is not None:
            ds = ds.isel(isel)
        if sel is not None:
            ds = ds.sel(sel)
        return ds

    def new_point_data(self, points, states_indices=None, n_states=None):
        """
        Creates a point data xarray object, containing only points.

        Parameters
        ----------
        points: numpy.ndarray
            The points, shape: (n_states, n_points, 3)
        states_indices: array_like, optional
            The indices of the states dimension
        n_states: int, optional
            The number of states

        Returns
        -------
        xarray.Dataset
            A dataset containing the points data

        """
        if n_states is None:
            n_states = self.n_states
        if states_indices is None:
            idata = {"coords": {}, "data_vars": {}}
        else:
            idata = {"coords": {FC.STATE: states_indices}, "data_vars": {}}

        if len(points.shape) == 2 and points.shape[1] == 3:
            pts = np.zeros((n_states,) + points.shape, dtype=config.dtype_double)
            pts[:] = points[None]
            points = pts
            del pts

        if (
            len(points.shape) != 3
            or points.shape[0] != n_states
            or points.shape[2] != 3
        ):
            raise ValueError(
                f"points have wrong dimensions, expecting ({n_states}, {points.shape[1]}, 3), got {points.shape}"
            )
        idata["data_vars"][FC.TARGETS] = (
            (FC.STATE, FC.TARGET, FC.TPOINT, FC.XYH),
            points[:, :, None, :],
        )
        idata["data_vars"][FC.TWEIGHTS] = (
            (FC.TPOINT,),
            np.array([1.0], dtype=config.dtype_double),
        )

        return xr.Dataset(**idata)

    def add_to_chunk_store(
        self,
        name,
        data,
        dims,
        mdata,
        tdata=None,
        copy=True,
    ):
        """
        Add data to the chunk store

        Parameters
        ----------
        name: str
            The data name
        data: numpy.ndarray
            The data
        dims: tuple
            The data dimensions
        mdata: foxes.core.MData
            The mdata object
        tdata: foxes.core.TData, optional
            The tdata object
        copy: bool
            Flag for copying incoming data

        """
        assert mdata.chunki_states is not None, (
            f"{self.name}: mdata.chunki_states is None, cannot add to chunk store"
        )
        assert mdata.chunki_points is not None, (
            f"{self.name}: mdata.chunki_points is None, cannot add to chunk store"
        )

        key = (mdata.chunki_states, mdata.chunki_points)
        if key not in self.chunk_store:
            n_states = int(mdata.n_states)
            n_targets = int(tdata.n_targets if tdata is not None else 0)
            self.chunk_store[key] = Dict(
                {
                    "i0": mdata.states_i0(counter=True),
                    "t0": tdata.targets_i0() if tdata is not None else 0,
                    "n_states": n_states,
                    "n_targets": n_targets,
                    "states_index": mdata[FC.STATE].copy(),
                    "dims": {"states_index": (FC.STATE,)},
                },
                _name=f"chunk_store_{key[0]}_{key[1]}",
            )

        self.chunk_store[key][name] = data.copy() if copy else data
        self.chunk_store[key]["dims"][name] = dims

    def get_from_chunk_store(
        self,
        name,
        mdata,
        prev_s=0,
        prev_t=0,
        ret_inds=False,
        error=True,
    ):
        """
        Get data to the chunk store

        Parameters
        ----------
        name: str
            The data name
        mdata: foxes.core.MData
            The mdata object
        tdata: foxes.core.TData, optional
            The tdata object
        prev_s: int
            How many states chunks backward
        prev_t: int
            How many points chunks backward
        ret_inds: bool
            Also return chunk index data (i0, n_states, t0, n_targets)
        error: bool
            Flag for raising KeyError if data not found

        Returns
        -------
        data: numpy.ndarray
            The data
        inds: tuple, optional
            The chunk index data (i0, n_states, t0, n_targets)

        """
        assert mdata.chunki_states is not None, (
            f"{self.name}: mdata.chunki_states is None, cannot add to chunk store"
        )
        assert mdata.chunki_points is not None, (
            f"{self.name}: mdata.chunki_points is None, cannot add to chunk store"
        )
        key = (mdata.chunki_states - prev_s, mdata.chunki_points - prev_t)

        try:
            chunk_data = self.chunk_store[key]
        except KeyError as e:
            if error:
                raise e
            else:
                return (None, (None, None, None, None)) if ret_inds else None

        chunk_states = chunk_data["states_index"]
        n_states = len(chunk_states)

        if (
            prev_s != 0
            or prev_t != 0
            or name not in chunk_data["dims"]
            or FC.STATE not in chunk_data["dims"][name]
            or (n_states == mdata.n_states and np.all(chunk_states == mdata[FC.STATE]))
        ):
            try:
                data = chunk_data[name]
            except KeyError as e:
                if error:
                    raise e
                else:
                    data = None

        # combine data from multiple chunks, in case of states subset selection:
        else:
            data = None
            for (__, ipoints), d in self.chunk_store.items():
                if ipoints == key[1] and name in d:
                    __, j0, j1 = np.intersect1d(
                        d["states_index"], mdata[FC.STATE], return_indices=True
                    )
                    if isinstance(d[name], dict):
                        if data is None:
                            data = {
                                k: np.full((mdata.n_states,) + v.shape[1:], np.nan)
                                for k, v in d[name].items()
                            }
                        for k in d[name]:
                            data[k][j1] = d[name][k][j0]
                    else:
                        if data is None:
                            data = np.full(
                                (mdata.n_states,) + d[name].shape[1:], np.nan
                            )
                        data[j1] = d[name][j0]

        if ret_inds:
            inds = (
                chunk_data["i0"],
                chunk_data["n_states"],
                chunk_data["t0"],
                chunk_data["n_targets"],
            )
            return data, inds
        else:
            return data

    def reset_chunk_store(self, new_chunk_store=None):
        """
        Resets the chunk store

        Parameters
        ----------
        new_chunk_store: foxes.utils.Dict, optional
            The new chunk store

        Returns
        -------
        chunk_store: foxes.utils.Dict
            The chunk store before resetting

        """
        chunk_store = self.chunk_store
        if new_chunk_store is None:
            self.__chunk_store = Dict(_name="chunk_store")
        elif isinstance(new_chunk_store, Dict):
            self.__chunk_store = new_chunk_store
        else:
            self.__chunk_store = Dict(_name="chunk_store")
            self.__chunk_store.update(new_chunk_store)
        return chunk_store

    def block_convergence(self, **kwargs):
        """
        Switch on convergence block during iterative run

        Parameters
        ----------
        kwargs: dict, optional
            Parameters for add_to_chunk_store()

        """
        self.add_to_chunk_store(
            name=FC.BLOCK_CONVERGENCE, data=True, dims=(), copy=False, **kwargs
        )

    def eval_conv_block(self):
        """
        Evaluate convergence block, removing blocks on the fly

        Returns
        -------
        blocked: bool
            True if convergence is currently blocked

        """
        blocked = False
        for c in self.__chunk_store.values():
            blocked = c.pop(FC.BLOCK_CONVERGENCE, False) or blocked
        return blocked

    def set_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to running, and moves
        all large data to stash.

        The stashed data will be returned by the
        unset_running() function after running calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Large data stash, this function adds data here, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        assert algo is self

        super().set_running(algo, data_stash, sel, isel, verbosity=verbosity)

        if data_stash is not None:
            data_stash[self.name].update(
                dict(
                    mbook=self.__mbook,
                    dbook=self.__dbook,
                    idata_mem=self.__idata_mem,
                )
            )
        del self.__mbook, self.__dbook
        self.__idata_mem = {}

    def unset_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to not running, recovering large data
        from stash

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Reconstruct model data from this stash, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        assert algo is self

        super().unset_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None:
            data = data_stash[self.name]
            self.__mbook = data.pop("mbook")
            self.__dbook = data.pop("dbook")
            self.__idata_mem = data.pop("idata_mem")
        else:
            self.reset_chunk_store()

    @abstractmethod
    def _launch_parallel_farm_calc(
        self,
        *args,
        mbook,
        dbook,
        chunk_store,
        **kwargs,
    ):
        """
        Runs the main farm calculation, launching parallelization

        Parameters
        ----------
        args: tuple, optional
            Additional parameters for running
        mbook: foxes.models.ModelBook
            The model book
        dbook: foxes.DataBook
            The data book, or None for default
        chunk_store: foxes.utils.Dict
            The chunk store
        kwargs: dict, optional
            Additional parameters for running

        Returns
        -------
        farm_results: xarray.Dataset
            The farm results. The calculated variables have
            dimensions (state, turbine)

        """
        pass

    def calc_farm(self, *args, clear_mem=False, **kwargs):
        """
        Calculate farm data.

        Parameters
        ----------
        args: tuple, optional
            Parameters
        clear_mem: bool
            Clear idata memory after starting the run
        kwargs: dict, optional
            Keyword parameters

        Returns
        -------
        farm_results: xarray.Dataset
            The farm results. The calculated variables have
            dimensions (state, turbine)

        """
        if self.running:
            raise ValueError(
                f"Algorithm '{self.name}': Cannot call calc_farm while running"
            )

        # set to running:
        data_stash = {} if not clear_mem else None
        chunk_store = self.reset_chunk_store()
        mdls = [
            m
            for m in [self] + list(args) + list(kwargs.values())
            if isinstance(m, Model)
        ]
        for m in mdls:
            m.set_running(
                self, data_stash, sel=None, isel=None, verbosity=self.verbosity - 2
            )

        # run parallel calculation:
        farm_results = self._launch_parallel_farm_calc(
            *args,
            chunk_store=chunk_store,
            # sel=None,
            # isel=None,
            **kwargs,
        )

        # reset to not running:
        for m in mdls:
            m.unset_running(
                self, data_stash, sel=None, isel=None, verbosity=self.verbosity - 2
            )

        return farm_results

    @abstractmethod
    def _launch_parallel_points_calc(
        self,
        *args,
        chunk_store,
        **kwargs,
    ):
        """
        Runs the main points calculation, launching parallelization

        Parameters
        ----------
        args: tuple, optional
            Additional parameters for running
        chunk_store: foxes.utils.Dict
            The chunk store
        kwargs: dict, optional
            Additional parameters for running

        Returns
        -------
        point_results: xarray.Dataset
            The point results. The calculated variables have
            dimensions (state, point)

        """
        pass

    def calc_points(self, *args, sel=None, isel=None, **kwargs):
        """
        Calculate points data.

        Parameters
        ----------
        args: tuple, optional
            Parameters
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        kwargs: dict, optional
            Keyword parameters

        Returns
        -------
        point_results: xarray.Dataset
            The point results. The calculated variables have
            dimensions (state, point)

        """
        if self.running:
            raise ValueError(
                f"Algorithm '{self.name}': Cannot call calc_points while running"
            )

        # set to running:
        data_stash = {}
        self.set_running(
            self, data_stash, sel=sel, isel=isel, verbosity=self.verbosity - 2
        )

        # run parallel calculation:
        chunk_store = self.reset_chunk_store()
        point_results = self._launch_parallel_points_calc(
            *args,
            chunk_store=chunk_store,
            sel=sel,
            isel=isel,
            **kwargs,
        )
        self.reset_chunk_store(chunk_store)

        # reset to not running:
        self.unset_running(
            self, data_stash, sel=sel, isel=isel, verbosity=self.verbosity - 2
        )

        return point_results

    def finalize(self, clear_mem=False):
        """
        Finalizes the algorithm.

        Parameters
        ----------
        clear_mem: bool
            Clear idata memory

        """
        if self.running:
            raise ValueError(f"Algorithm '{self.name}': Cannot finalize while running")
        super().finalize(self, self.verbosity - 1)
        if clear_mem:
            self.__idata_mem = Dict()
            # self.reset_chunk_store()

    @classmethod
    def new(cls, algo_type, *args, **kwargs):
        """
        Run-time algorithm factory.

        Parameters
        ----------
        algo_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for the constructor
        kwargs: dict, optional
            Additional parameters for the constructor

        """
        return new_instance(cls, algo_type, *args, **kwargs)



import numpy as np

import foxes.variables as FV
from foxes.config import config

from .rotor_points import RotorPoints


class PartialCentre(RotorPoints):
    """
    Partial wakes calculated only at the
    rotor centre point.

    :group: models.partial_wakes

    """

    def get_wake_points(self, algo, mdata, fdata):
        """
        Get the wake calculation points, and their
        weights.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        rpoints: numpy.ndarray
            The wake calculation points, shape:
            (n_states, n_turbines, n_tpoints, 3)
        rweights: numpy.ndarray
            The target point weights, shape: (n_tpoints,)

        """
        return fdata[FV.TXYH][:, :, None], np.ones(1, dtype=config.dtype_double)

    def map_rotor_results(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        variable,
        rotor_res,
        rotor_weights,
    ):
        """
        Map ambient rotor point results onto target points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        variable: str
            The variable name to map
        rotor_res: numpy.ndarray
            The results at rotor points, shape:
            (n_states, n_turbines, n_rotor_points)
        rotor_weights: numpy.ndarray
            The rotor point weights, shape: (n_rotor_points,)

        Returns
        -------
        res: numpy.ndarray
            The mapped results at target points, shape:
            (n_states, n_targets, n_tpoints)

        """
        if rotor_res.shape[2] > 1:
            return np.einsum(
                "str,r->st",
                rotor_res,
                rotor_weights,
            )[:, :, None]
        else:
            return rotor_res



import time
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--layout",
        help="The wind farm layout file (path or static)",
        default="test_farm_67.csv",
    )
    parser.add_argument(
        "-s",
        "--states",
        help="The states input file (path or static)",
        default="winds100.tab",
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default={}, nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["CrespoHernandez_quadratic", "Bastankhah2014_linear"],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument(
        "-sl",
        "--show_layout",
        help="Flag for showing layout figure",
        action="store_true",
    )
    parser.add_argument(
        "-cm", "--calc_mean", help="Calculate states mean", action="store_true"
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=5000,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    states = foxes.input.states.TabStates(
        data_source=args.states,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        fixed_vars={FV.RHO: 1.225, FV.TI: 0.05},
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        if not args.nofig:
            o = foxes.output.StatesRosePlotOutput(states, point=[0.0, 0.0, 100.0])
            o.get_figure(12, FV.AMB_WS, [0, 3.5, 6, 10, 15, 20])
            plt.show()

        farm = foxes.WindFarm()
        foxes.input.farm_layout.add_from_file(
            farm,
            args.layout,
            col_x="x",
            col_y="y",
            col_H="H",
            turbine_models=[ttype.name, "kTI_02"] + args.tmodels,
        )

        if not args.nofig and args.show_layout:
            ax = foxes.output.FarmLayoutOutput(farm).get_figure()
            plt.show()
            plt.close(ax.get_figure())

        algo = foxes.algorithms.Downwind(
            farm,
            states,
            mbook=mbook,
            rotor_model=args.rotor,
            wake_models=args.wakes,
            wake_frame="rotor_wd",
            partial_wakes=args.pwakes,
        )

        time0 = time.time()
        farm_results = algo.calc_farm()
        time1 = time.time()

        print("\nCalc time =", time1 - time0, "\n")

        print(farm_results)

        fr = farm_results.to_dataframe()
        print(fr[[FV.WD, FV.H, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P, FV.WEIGHT]])

        o = foxes.output.FarmResultsEval(farm_results)
        P0 = o.calc_mean_farm_power(ambient=True)
        P = o.calc_mean_farm_power()
        print(f"\nFarm power        : {P / 1000:.1f} MW")
        print(f"Farm ambient power: {P0 / 1000:.1f} MW")
        print(f"Farm efficiency   : {o.calc_farm_efficiency() * 100:.2f} %")
        print(f"Annual farm yield : {o.calc_farm_yield(algo=algo):.2f} GWh")

        if not args.nofig and args.calc_mean:
            o = foxes.output.FlowPlots2D(algo, farm_results)
            fig = o.get_mean_fig_xy(FV.WS, resolution=30)
            plt.show()
